# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2019 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test location related views."""

import pytest
from flask import json, url_for
from testutils import login_user

from invenio_files_rest.models import Bucket


def get_json(resp):
    """Get JSON from response."""
    return json.loads(resp.get_data(as_text=True))


@pytest.mark.parametrize(
    "user, expected",
    [
        (None, 401),
        ("auth", 403),
        ("location", 200),
    ],
)
def test_post_bucket(app, client, headers, dummy_location, permissions, user, expected):
    """Test post a bucket."""
    expected_keys = [
        "id",
        "links",
        "size",
        "quota_size",
        "max_file_size",
        "locked",
        "created",
        "updated",
    ]
    params = [{}, {"location_name": dummy_location.name}]

    login_user(client, permissions[user])

    for data in params:
        resp = client.post(
            url_for("invenio_files_rest.location_api"), data=data, headers=headers
        )
        assert resp.status_code == expected
        if resp.status_code == 200:
            resp_json = get_json(resp)
            for key in expected_keys:
                assert key in resp_json
            assert Bucket.get(resp_json["id"])


@pytest.mark.parametrize(
    "user, expected",
    [
        (None, 405),
        ("auth", 405),
        ("location", 405),
    ],
)
def test_get_location(
    app, client, headers, dummy_location, permissions, user, expected
):
    """Test GET a location."""
    login_user(client, permissions[user])
    r = client.get(url_for("invenio_files_rest.location_api"), headers=headers)
    assert r.status_code == expected



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2019 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test file signals."""

from io import BytesIO

from flask import url_for
from testutils import login_user

from invenio_files_rest.signals import file_deleted, file_uploaded


def test_signals(app, client, headers, bucket, permissions):
    """Test file_uploaded and file_deleted signals."""
    login_user(client, permissions["bucket"])
    key = "myfile.txt"
    data = b"content of my file"
    object_url = url_for("invenio_files_rest.object_api", bucket_id=bucket.id, key=key)

    calls = []

    def upload_listener(sender, obj=None):
        calls.append("file-uploaded")

    def delete_listener(sender, obj=None):
        calls.append("file-deleted")

    file_uploaded.connect(upload_listener, weak=False)
    file_deleted.connect(delete_listener, weak=False)
    try:
        client.put(
            object_url,
            input_stream=BytesIO(data),
            headers={"Content-Type": "application/octet-stream"},
        )
        client.delete(object_url)
        assert calls == ["file-uploaded", "file-deleted"]
    finally:
        file_uploaded.disconnect(upload_listener)
        file_deleted.disconnect(delete_listener)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2019 CERN.
# Copyright (C) 2023 Graz University of Technology.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.


"""Pytest configuration."""

import hashlib
import os
import shutil
import tempfile
from io import BytesIO

import pytest
from flask import Flask, json, url_for
from flask_celeryext import FlaskCeleryExt
from flask_menu import Menu
from invenio_access import InvenioAccess
from invenio_access.models import ActionRoles, ActionUsers, Role
from invenio_accounts import InvenioAccounts
from invenio_accounts.testutils import create_test_user
from invenio_accounts.views import blueprint as accounts_blueprint
from invenio_db import InvenioDB
from invenio_db import db as db_
from invenio_db.utils import drop_alembic_version_table
from invenio_i18n import Babel, InvenioI18N
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.schema import DropConstraint, DropSequence, DropTable
from sqlalchemy_utils.functions import create_database, database_exists

from invenio_files_rest import InvenioFilesREST
from invenio_files_rest.models import (
    Bucket,
    Location,
    MultipartObject,
    ObjectVersion,
    Part,
)
from invenio_files_rest.permissions import (
    bucket_listmultiparts_all,
    bucket_read_all,
    bucket_read_versions_all,
    bucket_update_all,
    location_update_all,
    multipart_delete_all,
    multipart_read_all,
    object_delete_all,
    object_delete_version_all,
    object_read_all,
    object_read_version_all,
)
from invenio_files_rest.storage import PyFSFileStorage
from invenio_files_rest.views import blueprint


@compiles(DropTable, "postgresql")
def _compile_drop_table(element, compiler, **kwargs):
    return compiler.visit_drop_table(element) + " CASCADE"


@compiles(DropConstraint, "postgresql")
def _compile_drop_constraint(element, compiler, **kwargs):
    return compiler.visit_drop_constraint(element) + " CASCADE"


@compiles(DropSequence, "postgresql")
def _compile_drop_sequence(element, compiler, **kwargs):
    return compiler.visit_drop_sequence(element) + " CASCADE"


@pytest.fixture()
def base_app():
    """Flask application fixture."""
    app_ = Flask("testapp")
    app_.config.update(
        TESTING=True,
        # Celery 3
        CELERY_ALWAYS_EAGER=True,
        CELERY_EAGER_PROPAGATES_EXCEPTIONS=True,
        # Celery 4
        CELERY_TASK_ALWAYS_EAGER=True,
        CELERY_TASK_EAGER_PROPAGATES=True,
        SQLALCHEMY_TRACK_MODIFICATIONS=True,
        SQLALCHEMY_DATABASE_URI=os.environ.get(
            "SQLALCHEMY_DATABASE_URI", "sqlite:///:memory:"
        ),
        WTF_CSRF_ENABLED=False,
        SERVER_NAME="invenio.org",
        SECURITY_PASSWORD_SALT="TEST_SECURITY_PASSWORD_SALT",
        SECRET_KEY="TEST_SECRET_KEY",
        FILES_REST_MULTIPART_CHUNKSIZE_MIN=2,
        FILES_REST_MULTIPART_CHUNKSIZE_MAX=20,
        FILES_REST_MULTIPART_MAX_PARTS=100,
        FILES_REST_TASK_WAIT_INTERVAL=0.1,
        FILES_REST_TASK_WAIT_MAX_SECONDS=1,
        APP_THEME=["semantic-ui"],
        THEME_ICONS={"semantic-ui": dict(link="linkify icon")},
    )

    FlaskCeleryExt(app_)
    InvenioDB(app_)
    Babel(app_)
    Menu(app_)

    return app_


@pytest.yield_fixture()
def app(base_app):
    """Flask application fixture."""
    InvenioI18N(base_app)
    InvenioAccounts(base_app)
    InvenioAccess(base_app)
    base_app.register_blueprint(accounts_blueprint)
    InvenioFilesREST(base_app)
    base_app.register_blueprint(blueprint)

    with base_app.app_context():
        yield base_app


@pytest.yield_fixture()
def db(app):
    """Get setup database."""
    if not database_exists(str(db_.engine.url)):
        create_database(str(db_.engine.url))
    db_.create_all()
    yield db_
    db_.session.remove()
    db_.drop_all()
    drop_alembic_version_table()


@pytest.yield_fixture()
def client(app):
    """Get test client."""
    with app.test_client() as client:
        yield client


@pytest.yield_fixture()
def dummy_location(db):
    """File system location."""
    tmppath = tempfile.mkdtemp()

    loc = Location(name="testloc", uri=tmppath, default=True)
    db.session.add(loc)
    db.session.commit()

    yield loc

    shutil.rmtree(tmppath)


@pytest.fixture()
def pyfs_testpath(dummy_location):
    """Temporary path for PyFS."""
    return os.path.join(dummy_location.uri, "subpath/data")


@pytest.fixture()
def pyfs(dummy_location, pyfs_testpath):
    """Instance of PyFSFileStorage."""
    return PyFSFileStorage(pyfs_testpath)


@pytest.yield_fixture()
def extra_location(db):
    """File system location."""
    tmppath = tempfile.mkdtemp()

    loc = Location(name="extra", uri=tmppath, default=False)
    db.session.add(loc)
    db.session.commit()

    yield loc

    shutil.rmtree(tmppath)


@pytest.fixture()
def bucket(db, dummy_location):
    """File system location."""
    b1 = Bucket.create()
    db.session.commit()
    return b1


@pytest.fixture()
def multipart(db, bucket):
    """Multipart object."""
    mp = MultipartObject.create(bucket, "mykey", 110, 20)
    db.session.commit()
    return mp


@pytest.fixture()
def multipart_url(multipart):
    """File system location."""
    return url_for(
        "invenio_files_rest.object_api",
        bucket_id=str(multipart.bucket_id),
        key=multipart.key,
        uploadId=multipart.upload_id,
    )


@pytest.fixture()
def parts(db, multipart):
    """All parts for a multipart object."""
    items = []
    for i in range(multipart.last_part_number + 1):
        chunk_size = (
            multipart.chunk_size
            if not i == multipart.last_part_number
            else multipart.last_part_size
        )
        p = Part.create(
            multipart, i, stream=BytesIO("{0}".format(i).encode("ascii") * chunk_size)
        )
        items.append(p)

    db.session.commit()
    return items


@pytest.yield_fixture()
def objects(db, bucket):
    """File system location."""
    # Create older versions first
    for key, content in [("LICENSE", b"old license"), ("README.rst", b"old readme")]:
        ObjectVersion.create(bucket, key, stream=BytesIO(content), size=len(content))

    # Create new versions
    objs = []
    for key, content in [("LICENSE", b"license file"), ("README.rst", b"readme file")]:
        objs.append(
            ObjectVersion.create(
                bucket, key, stream=BytesIO(content), size=len(content)
            )
        )
    db.session.commit()

    yield objs


@pytest.yield_fixture()
def versions(objects):
    """Get objects with all their versions."""
    versions = []
    for obj in objects:
        versions.extend(ObjectVersion.get_versions(obj.bucket, obj.key))

    yield versions


@pytest.fixture()
def users_data(db):
    """User data fixture."""
    return [
        dict(email="user1@inveniosoftware.org", password="pass1"),
        dict(email="user2@inveniosoftware.org", password="pass1"),
    ]


@pytest.fixture()
def users(db, users_data):
    """Create test users."""
    return [
        create_test_user(active=True, **users_data[0]),
        create_test_user(active=True, **users_data[1]),
    ]


@pytest.fixture()
def headers():
    """Get standard Invenio REST API headers."""
    return {
        "Content-Type": "application/json",
        "Accept": "*/*",
    }


@pytest.yield_fixture()
def permissions(db, bucket):
    """Permission for users."""
    users = {
        None: None,
    }

    for user in ["auth", "location", "bucket", "objects", "objects-read-version"]:
        users[user] = create_test_user(
            email="{0}@invenio-software.org".format(user), password="pass1", active=True
        )

    location_perms = [
        location_update_all,
        bucket_read_all,
        bucket_read_versions_all,
        bucket_update_all,
        bucket_listmultiparts_all,
        object_read_all,
        object_read_version_all,
        object_delete_all,
        object_delete_version_all,
        multipart_read_all,
        multipart_delete_all,
    ]

    bucket_perms = [
        bucket_read_all,
        object_read_all,
        bucket_update_all,
        object_delete_all,
        multipart_read_all,
    ]

    objects_perms = [
        object_read_all,
    ]

    for perm in location_perms:
        db.session.add(ActionUsers(action=perm.value, user=users["location"]))
    for perm in bucket_perms:
        db.session.add(
            ActionUsers(
                action=perm.value, argument=str(bucket.id), user=users["bucket"]
            )
        )
    for perm in objects_perms:
        db.session.add(
            ActionUsers(
                action=perm.value, argument=str(bucket.id), user=users["objects"]
            )
        )
    db.session.commit()

    yield users


@pytest.yield_fixture()
def admin_user(db):
    """Permission for admin users."""
    perms = [
        location_update_all,
        bucket_read_all,
        bucket_read_versions_all,
        bucket_update_all,
        bucket_listmultiparts_all,
        object_read_all,
        object_read_version_all,
        object_delete_all,
        object_delete_version_all,
        multipart_read_all,
        multipart_delete_all,
        bucket_read_all,
        object_read_all,
        bucket_update_all,
        object_delete_all,
        multipart_read_all,
        object_read_all,
    ]

    admin = Role(name="admin")

    for perm in perms:
        db.session.add(ActionRoles.allow(perm, role=admin))

    admin_user = create_test_user(
        email="admin@invenio-software.org", password="pass1", active=True
    )
    admin.users.append(admin_user)

    db.session.commit()

    yield admin_user


@pytest.fixture()
def get_md5():
    """Get MD5 of data."""

    def inner(data, prefix=True):
        m = hashlib.md5()
        m.update(data)
        return "md5:{0}".format(m.hexdigest()) if prefix else m.hexdigest()

    return inner


@pytest.fixture()
def get_json():
    """Get JSON from response."""

    def inner(resp, code=None):
        if code is not None:
            assert resp.status_code == code
        return json.loads(resp.get_data(as_text=True))

    return inner


@pytest.fixture()
def offload_file_serving(app):
    """Serve a redirect instead of streaming the file."""
    app.config["FILES_REST_XSENDFILE_ENABLED"] = True
    yield app
    app.config["FILES_REST_XSENDFILE_ENABLED"] = False



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2019 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Module test views."""

import errno
from io import BytesIO
from os.path import exists, join
from unittest.mock import MagicMock, patch

import pytest
from fs.errors import FSError, ResourceNotFound

from invenio_files_rest.models import Bucket, FileInstance, ObjectVersion
from invenio_files_rest.tasks import (
    clear_orphaned_files,
    migrate_file,
    remove_file_data,
    schedule_checksum_verification,
    verify_checksum,
)


def test_verify_checksum(app, db, dummy_location):
    """Test celery tasks for checksum verification."""
    b1 = Bucket.create()
    with open("README.rst", "rb") as fp:
        obj = ObjectVersion.create(b1, "README.rst", stream=fp)
    db.session.commit()
    file_id = obj.file_id

    verify_checksum(str(file_id))

    f = FileInstance.query.get(file_id)
    assert f.last_check_at
    assert f.last_check is True

    f.uri = "invalid"
    db.session.add(f)
    db.session.commit()
    pytest.raises(ResourceNotFound, verify_checksum, str(file_id), throws=True)

    f = FileInstance.query.get(file_id)
    assert f.last_check is True

    verify_checksum(str(file_id), throws=False)
    f = FileInstance.query.get(file_id)
    assert f.last_check is None

    f.last_check = True
    db.session.add(f)
    db.session.commit()
    with pytest.raises(ResourceNotFound):
        verify_checksum(str(file_id), pessimistic=True)
    f = FileInstance.query.get(file_id)
    assert f.last_check is None


def test_schedule_checksum_verification(app, db, dummy_location):
    """Test file checksum verification scheduling celery task."""
    b1 = Bucket.create()
    objects = []
    for i in range(100):
        objects.append(ObjectVersion.create(b1, str(i), stream=BytesIO(b"tests")))
    db.session.commit()

    # 100 files of the 5-byte content 'tests' should be 500 bytes in total
    assert sum(o.file.size for o in objects) == 500

    for obj in objects:
        assert obj.file.last_check is True
        assert obj.file.last_check_at is None

    schedule_task = schedule_checksum_verification.s(
        frequency={"minutes": 20}, batch_interval={"minutes": 1}
    )

    def checked_files():
        return len([o for o in ObjectVersion.get_by_bucket(b1) if o.file.last_check_at])

    # Scheduling for 100 files, for all of them to be checked every 20 minutes,
    # with batches of equal number of file being sent out every minute
    # should total to 20 batches with 5 files per batch.
    schedule_task.apply(kwargs={"max_count": 0})
    assert checked_files() == 5

    # Repeat the schedule
    schedule_task.apply(kwargs={"max_count": 0})
    assert checked_files() == 10

    schedule_task.apply(kwargs={"max_count": 3})  # 3 files are checked
    assert checked_files() == 13

    # Scheduling for 500 bytes of files, for all of them to be checked every 20
    # minutes, with equal in size batches being sent out every minute should
    # total to 20 batches of 25 bytes per batch (5 files).
    schedule_task.apply(kwargs={"max_size": 0})
    assert checked_files() == 18

    schedule_task.apply(kwargs={"max_size": 15})  # 3 files are checked
    assert checked_files() == 21


def test_migrate_file(app, db, dummy_location, extra_location, bucket, objects):
    """Test file migration."""
    obj = objects[0]

    # Test pre-condition
    old_uri = obj.file.uri
    assert exists(old_uri)
    assert old_uri == join(
        dummy_location.uri,
        str(obj.file.id)[0:2],
        str(obj.file.id)[2:4],
        str(obj.file.id)[4:],
        "data",
    )
    assert FileInstance.query.count() == 4

    # Migrate file
    with patch("invenio_files_rest.tasks.verify_checksum") as verify_checksum:
        migrate_file(
            obj.file_id, location_name=extra_location.name, post_fixity_check=True
        )
        assert verify_checksum.delay.called

    # Get object again
    obj = ObjectVersion.get(bucket, obj.key)
    new_uri = obj.file.uri
    assert exists(old_uri)
    assert exists(new_uri)
    assert new_uri != old_uri
    assert FileInstance.query.count() == 5


def test_migrate_file_copyfail(
    app, db, dummy_location, extra_location, bucket, objects
):
    """Test a failed copy."""
    obj = objects[0]

    assert FileInstance.query.count() == 4
    with patch("fs.osfs.io") as io:
        e = OSError()
        e.errno = errno.EPERM
        io.open = MagicMock(side_effect=e)
        pytest.raises(
            FSError, migrate_file, obj.file_id, location_name=extra_location.name
        )
    assert FileInstance.query.count() == 4


def test_remove_file_data(app, db, dummy_location, versions):
    """Test remove file data."""
    # Remove an object, so file instance have no references
    obj = versions[1]
    assert obj.is_head is False
    file_ = obj.file
    obj.remove()
    db.session.commit()

    # Remove the file instance - file not writable
    assert exists(file_.uri)
    assert FileInstance.query.count() == 4
    remove_file_data(str(file_.id))
    assert FileInstance.query.count() == 4
    assert exists(file_.uri)

    # Remove the file instance - file is writable
    file_.writable = True
    db.session.commit()
    assert exists(file_.uri)
    assert FileInstance.query.count() == 4
    remove_file_data(str(file_.id))
    assert FileInstance.query.count() == 3
    assert not exists(file_.uri)

    # Try to remove file instance with references.
    obj = versions[0]
    assert exists(obj.file.uri)
    assert FileInstance.query.count() == 3
    remove_file_data(str(obj.file.id))
    assert exists(obj.file.uri)


def test_clear_orphaned_files(app, db, dummy_location, versions):
    """Test clearing orphan files."""
    # create an orphaned file
    obj = versions[1]
    file_ = obj.file
    for obj in file_.objects:
        obj.remove()

    file_.writable = False
    db.session.commit()

    # make sure that the file is orphaned
    assert not file_.objects

    # try to delete the orphan which is marked as not writable
    assert exists(file_.uri)
    assert FileInstance.query.count() == 4
    clear_orphaned_files()
    assert exists(file_.uri)
    assert FileInstance.query.count() == 4

    # try to force the deletion
    assert exists(file_.uri)
    assert FileInstance.query.count() == 4
    clear_orphaned_files(force_delete_check=lambda fi: True)
    assert not exists(file_.uri)
    assert FileInstance.query.count() == 3



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2019 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""File size limiting functionality for Invenio-Files-REST."""


def file_size_limiters(bucket):
    """Get default file size limiters.

    :param bucket: The :class:`invenio_files_rest.models.Bucket` instance.
    :returns: A list containing an instance of
        :class:`invenio_files_rest.limiters.FileSizeLimit` with quota left
        value and description and another one with max file size value and
        description.
    """
    return [
        FileSizeLimit(
            bucket.quota_left,
            "Bucket quota exceeded.",
        ),
        FileSizeLimit(
            bucket.max_file_size,
            "Maximum file size exceeded.",
        ),
    ]


class FileSizeLimit(object):
    """File size limiter."""

    not_implemented_error = NotImplementedError(
        "FileSizeLimit supports only comparisons with integers and other "
        "FileSizeLimits."
    )

    def __init__(self, limit, reason):
        """Instantiate a new file size limit.

        :param limit: The imposed imposed limit.
        :param reason: The limit description.
        """
        self.limit = limit
        self.reason = reason

    def __lt__(self, other):
        """Check if this limit is less than the other one."""
        if isinstance(other, int):
            return self.limit < other
        elif isinstance(other, FileSizeLimit):
            return self.limit < other.limit
        raise self.not_implemented_error

    def __gt__(self, other):
        """Check if this limit is greater than the other one."""
        if isinstance(other, int):
            return self.limit > other
        elif isinstance(other, FileSizeLimit):
            return self.limit > other.limit
        raise self.not_implemented_error

    def __eq__(self, other):
        """Check for equality."""
        if isinstance(other, int):
            return self.limit == other
        elif isinstance(other, FileSizeLimit):
            return self.limit == other.limit
        raise self.not_implemented_error



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2019 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Celery tasks for Invenio-Files-REST."""

import math
import uuid
from datetime import date, datetime, timedelta

import sqlalchemy as sa
from celery import current_app as current_celery
from celery import current_task, group, shared_task
from celery.states import state
from celery.utils.log import get_task_logger
from flask import current_app
from invenio_db import db
from sqlalchemy.exc import IntegrityError

from .models import FileInstance, Location, MultipartObject, ObjectVersion
from .signals import file_uploaded
from .utils import obj_or_import_string

logger = get_task_logger(__name__)


def progress_updater(size, total):
    """Progress reporter for checksum verification."""
    current_task.update_state(
        state=state("PROGRESS"), meta=dict(size=size, total=total)
    )


@shared_task(ignore_result=True)
def verify_checksum(
    file_id, pessimistic=False, chunk_size=None, throws=True, checksum_kwargs=None
):
    """Verify checksum of a file instance.

    :param file_id: The file ID.
    """
    f = FileInstance.query.get(uuid.UUID(file_id))

    # Anything might happen during the task, so being pessimistic and marking
    # the file as unchecked is a reasonable precaution
    if pessimistic:
        f.clear_last_check()
        db.session.commit()
    f.verify_checksum(
        progress_callback=progress_updater,
        chunk_size=chunk_size,
        throws=throws,
        checksum_kwargs=checksum_kwargs,
    )
    db.session.commit()


def default_checksum_verification_files_query():
    """Return a query of valid FileInstances for checksum verification."""
    return FileInstance.query


@shared_task(ignore_result=True)
def schedule_checksum_verification(
    frequency=None,
    batch_interval=None,
    max_count=None,
    max_size=None,
    files_query=None,
    checksum_kwargs=None,
):
    """Schedule a batch of files for checksum verification.

    The purpose of this task is to be periodically called through `celerybeat`,
    in order achieve a repeated verification cycle of all file checksums, while
    following a set of constraints in order to throttle the execution rate of
    the checks.

    :param dict frequency: Time period over which a full check of all files
        should be performed. The argument is a dictionary that will be passed
        as arguments to the `datetime.timedelta` class. Defaults to a month (30
        days).
    :param dict batch_interval: How often a batch is sent. If not supplied,
        this information will be extracted, if possible, from the
        celery.conf['CELERYBEAT_SCHEDULE'] entry of this task. The argument is
        a dictionary that will be passed as arguments to the
        `datetime.timedelta` class.
    :param int max_count: Max count of files of a single batch. When set to `0`
        it's automatically calculated to be distributed equally through the
        number of total batches.
    :param int max_size: Max size of a single batch in bytes. When set to `0`
        it's automatically calculated to be distributed equally through the
        number of total batches.
    :param str files_query: Import path for a function returning a
        FileInstance query for files that should be checked.
    :param dict checksum_kwargs: Passed to ``FileInstance.verify_checksum``.
    """
    assert max_count is not None or max_size is not None
    frequency = timedelta(**frequency) if frequency else timedelta(days=30)
    if batch_interval:
        batch_interval = timedelta(**batch_interval)
    else:
        celery_schedule = current_celery.conf.get("CELERYBEAT_SCHEDULE", {})
        batch_interval = batch_interval or next(
            (
                v["schedule"]
                for v in celery_schedule.values()
                if v.get("task") == schedule_checksum_verification.name
            ),
            None,
        )
    if not batch_interval or not isinstance(batch_interval, timedelta):
        raise Exception('No "batch_interval" could be decided')

    total_batches = int(frequency.total_seconds() / batch_interval.total_seconds())

    files = obj_or_import_string(
        files_query, default=default_checksum_verification_files_query
    )()
    files = files.order_by(sa.func.coalesce(FileInstance.last_check_at, date.min))

    if max_count is not None:
        all_files_count = files.count()
        min_count = int(math.ceil(all_files_count / total_batches))
        max_count = min_count if max_count == 0 else max_count
        if max_count < min_count:
            current_app.logger.warning(
                'The "max_count" you specified ({0}) is smaller than the '
                "minimum batch file count required ({1}) in order to achieve "
                "the file checks over the specified period ({2}).".format(
                    max_count, min_count, frequency
                )
            )
        files = files.limit(max_count)

    if max_size is not None:
        all_files_size = db.session.query(sa.func.sum(FileInstance.size)).scalar()
        min_size = int(math.ceil(all_files_size / total_batches))
        max_size = min_size if max_size == 0 else max_size
        if max_size < min_size:
            current_app.logger.warning(
                'The "max_size" you specified ({0}) is smaller than the '
                "minimum batch total file size required ({1}) in order to "
                "achieve the file checks over the specified period ({2}).".format(
                    max_size, min_size, frequency
                )
            )

    files = files.yield_per(1000)
    scheduled_file_ids = []
    total_size = 0
    for f in files:
        # Add at least the first file, since it might be larger than "max_size"
        scheduled_file_ids.append(str(f.id))
        total_size += f.size
        if max_size and max_size <= total_size:
            break
    group(
        verify_checksum.s(
            file_id,
            pessimistic=True,
            throws=False,
            checksum_kwargs=(checksum_kwargs or {}),
        )
        for file_id in scheduled_file_ids
    ).apply_async()


@shared_task(ignore_result=True, max_retries=3, default_retry_delay=20 * 60)
def migrate_file(src_id, location_name, post_fixity_check=False):
    """Task to migrate a file instance to a new location.

    .. note:: If something goes wrong during the content copy, the destination
        file instance is removed.

    :param src_id: The :class:`invenio_files_rest.models.FileInstance` ID.
    :param location_name: Where to migrate the file.
    :param post_fixity_check: Verify checksum after migration.
        (Default: ``False``)
    """
    location = Location.get_by_name(location_name)
    f_src = FileInstance.get(src_id)

    # Create destination
    f_dst = FileInstance.create()
    db.session.commit()

    try:
        # Copy contents
        f_dst.copy_contents(
            f_src,
            progress_callback=progress_updater,
            default_location=location.uri,
        )
        db.session.commit()
    except Exception:
        # Remove destination file instance if an error occurred.
        db.session.delete(f_dst)
        db.session.commit()
        raise

    # Update all objects pointing to file.
    ObjectVersion.relink_all(f_src, f_dst)
    db.session.commit()

    # Start a fixity check
    if post_fixity_check:
        verify_checksum.delay(str(f_dst.id))


@shared_task(ignore_result=True)
def remove_file_data(file_id, silent=True, force=False):
    """Remove file instance and associated data.

    :param file_id: The :class:`invenio_files_rest.models.FileInstance` ID.
    :param silent: It stops propagation of a possible raised IntegrityError
        exception. (Default: ``True``)
    :param force: Whether to delete the file even if the file instance is not
        marked as writable.
    :raises sqlalchemy.exc.IntegrityError: Raised if the database removal goes
        wrong and silent is set to ``False``.
    """
    try:
        # First remove FileInstance from database and commit transaction to
        # ensure integrity constraints are checked and enforced.
        f = FileInstance.get(file_id)
        if not f.writable and not force:
            return
        f.delete()
        db.session.commit()
        # Next, remove the file on disk. This leaves the possibility of having
        # a file on disk dangling in case the database removal works, and the
        # disk file removal doesn't work.
        f.storage().delete()
    except IntegrityError:
        if not silent:
            raise


@shared_task()
def merge_multipartobject(upload_id, version_id=None):
    """Merge multipart object.

    :param upload_id: The :class:`invenio_files_rest.models.MultipartObject`
        upload ID.
    :param version_id: Optionally you can define which file version.
        (Default: ``None``)
    :returns: The :class:`invenio_files_rest.models.ObjectVersion` version
        ID.
    """
    mp = MultipartObject.query.filter_by(upload_id=upload_id).one_or_none()
    if not mp:
        raise RuntimeError("Upload ID does not exists.")
    if not mp.completed:
        raise RuntimeError("MultipartObject is not completed.")

    try:
        obj = mp.merge_parts(version_id=version_id, progress_callback=progress_updater)
        db.session.commit()
        file_uploaded.send(current_app._get_current_object(), obj=obj)
        return str(obj.version_id)
    except Exception:
        db.session.rollback()
        raise


@shared_task(ignore_result=True)
def remove_expired_multipartobjects():
    """Remove expired multipart objects."""
    delta = current_app.config["FILES_REST_MULTIPART_EXPIRES"]
    expired_dt = datetime.utcnow() - delta

    file_ids = []
    for mp in MultipartObject.query_expired(expired_dt):
        file_ids.append(str(mp.file_id))
        mp.delete()

    for fid in file_ids:
        remove_file_data.delay(fid)


@shared_task(ignore_result=True)
def clear_orphaned_files(force_delete_check=lambda file_instance: False, limit=1000):
    """Delete orphaned files from DB and storage.

    .. note::

        Orphan files are files
        (:class:`invenio_files_rest.models.FileInstance` objects and their
        on-disk counterparts) that do not have any
        :class:`invenio_files_rest.models.ObjectVersion` objects associated
        with them (anymore).

    The celery beat configuration for scheduling this task may set values for
    this task's parameters:

    .. code-block:: python

        "clear-orphan-files": {
            "task": "invenio_files_rest.tasks.clear_orphaned_files",
            "schedule": 60 * 60 * 24,
            "kwargs": {
                "force_delete_check": lambda file: False,
                "limit": 500,
            }
        }

    :param force_delete_check: A function to be called on each orphan file instance
                               to check if its deletion should be forced (bypass the
                               check of its ``writable`` flag).
                               For example, this function can be used to force-delete
                               files only if they are located on the local file system.
                               Signature: The function should accept a
                               :class:`invenio_files_rest.models.FileInstance` object
                               and return a boolean value.
                               Default: Never force-delete any orphan files
                               (``lambda file_instance: False``).
    :param limit: Limit for the number of orphan files considered for deletion in each
                  task execution (and thus the number of generated celery tasks).
                  A value of zero (0) or lower disables the limit.
    """
    # with the tilde (~) operator, we get all file instances that *don't*
    # have any associated object versions
    query = FileInstance.query.filter(~FileInstance.objects.any())
    if limit > 0:
        query = query.limit(limit)

    for orphan in query:
        # note: the ID is cast to str because serialization of UUID objects can fail
        remove_file_data.delay(str(orphan.id), force=force_delete_check(orphan))



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2019 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Form data parser customization.

Flask and Werkzeug by default checks a request's content length against
``MAX_CONTENT_LENGTH`` configuration variable as soon as *any* content type is
specified in the request and not only when a form data content type is
specified. This behavior prevents both setting a MAX_CONTENT_LENGTH for form
data *and* allowing streaming uploads of large binary files.

In order to allow for both max content length checking and streaming uploads
of large files, we provide a custom form data parser which only checks the
content length if a form data content type is specified. The custom form data
parser is installed by using the custom Flask application class provided
provided in this module.
"""

from flask import Flask as FlaskBase

from .wrappers import Request


class Flask(FlaskBase):
    """Flask application class needed to use custom request class."""

    request_class = Request



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2019 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Models for Invenio-Files-REST.

The entities of this module consists of:

 * **Buckets** - Identified by UUIDs, and contains objects.
 * **Buckets tags** - Identified uniquely with a bucket by a key. Used to store
   extra metadata for a bucket.
 * **Objects** - Identified uniquely within a bucket by string keys. Each
   object can have multiple object versions (note: Objects do not have their
   own database table).
 * **Object versions** - Identified by UUIDs and belongs to one specific object
   in one bucket. Each object version has zero or one file instance. If the
   object version has no file instance, it is considered a *delete marker*.
 * **File instance** - Identified by UUIDs. Represents a physical file on disk.
   The location of the file is specified via a URI. A file instance can have
   many object versions.
 * **Locations** - A bucket belongs to a specific location. Locations can be
   used to represent e.g. different storage systems.
 * **Multipart Objects** - Identified by UUIDs and belongs to a specific bucket
   and key.
 * **Part object** - Identified by their multipart object and a part number.

The actual file access is handled by a storage interface. Also, objects do not
have their own model, but are represented via the :py:data:`ObjectVersion`
model.
"""

import re
import sys
import uuid
from datetime import datetime
from functools import wraps
from os.path import basename

from flask import current_app
from invenio_db import db
from sqlalchemy.dialects import mysql
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import validates
from sqlalchemy.orm.exc import MultipleResultsFound
from sqlalchemy_utils.types import UUIDType

from .errors import (
    BucketLockedError,
    FileInstanceAlreadySetError,
    FileInstanceUnreadableError,
    FileSizeError,
    InvalidKeyError,
    InvalidOperationError,
    MultipartAlreadyCompleted,
    MultipartInvalidChunkSize,
    MultipartInvalidPartNumber,
    MultipartInvalidSize,
    MultipartMissingParts,
    MultipartNotCompleted,
)
from .proxies import current_files_rest
from .utils import guess_mimetype

slug_pattern = re.compile("^[a-z][a-z0-9-]+$")


#
# Helpers
#
def validate_key(key):
    """Validate key.

    :param key: The key to validate.
    :raises invenio_files_rest.errors.InvalidKeyError: If the key is longer
        than the maximum length defined in
        :data:`invenio_files_rest.config.FILES_REST_FILE_URI_MAX_LEN`.
    :returns: The key.
    """
    if len(key) > current_app.config["FILES_REST_OBJECT_KEY_MAX_LEN"]:
        raise InvalidKeyError()
    return key


def as_bucket(value):
    """Get a bucket object from a bucket ID or a bucket object.

    :param value: A :class:`invenio_files_rest.models.Bucket` or a Bucket ID.
    :returns: A :class:`invenio_files_rest.models.Bucket` instance.
    """
    return value if isinstance(value, Bucket) else Bucket.get(value)


def as_bucket_id(value):
    """Get a bucket ID from a bucket ID or a bucket object.

    :param value: A :class:`invenio_files_rest.models.Bucket` instance of a
        bucket ID.
    :returns: The :class:`invenio_files_rest.models.Bucket` ID.
    """
    return value.id if isinstance(value, Bucket) else value


def as_object_version(value):
    """Get an object version object from an object version ID or an object version.

    :param value: A :class:`invenio_files_rest.models.ObjectVersion` or an
        object version ID.
    :returns: A :class:`invenio_files_rest.models.ObjectVersion` instance.
    """
    return (
        value
        if isinstance(value, ObjectVersion)
        else ObjectVersion.query.filter_by(version_id=value).one_or_none()
    )


def as_object_version_id(value):
    """Get an object version ID from an object version ID or an object version.

    :param value: A :class:`invenio_files_rest.models.ObjectVersion` instance
        of a object version ID.
    :returns: The :class:`invenio_files_rest.models.ObjectVersion` version_id.
    """
    return value.version_id if isinstance(value, ObjectVersion) else value


#
# Decorators to validate state.
#
def update_bucket_size(f):
    """Decorate to update bucket size after operation."""

    @wraps(f)
    def inner(self, *args, **kwargs):
        res = f(self, *args, **kwargs)
        self.bucket.size += self.file.size
        return res

    return inner


def ensure_state(default_getter, exc_class, default_msg=None):
    """Create a decorator factory function."""

    def decorator(getter=default_getter, msg=default_msg):
        def ensure_decorator(f):
            @wraps(f)
            def inner(self, *args, **kwargs):
                if not getter(self):
                    raise exc_class(msg) if msg else exc_class()
                return f(self, *args, **kwargs)

            return inner

        return ensure_decorator

    return decorator


class BucketError(object):
    """Represents a bucket level error.

    .. note:: This is not an actual exception.
    """

    def __init__(self, message):
        self.res = dict(message=message)

    def to_dict(self):
        return self.res


class ObjectVersionError(object):
    """Represents an object version level error.

    .. note:: This is not an actual exception.
    """

    def __init__(self, message):
        self.res = dict(message=message)

    def to_dict(self):
        return self.res


ensure_readable = ensure_state(lambda o: o.readable, FileInstanceUnreadableError)
"""Ensure file is readable."""

ensure_writable = ensure_state(
    lambda o: o.writable, ValueError, "File is not writable."
)
"""Ensure file is writeable."""

ensure_completed = ensure_state(lambda o: o.completed, MultipartNotCompleted)
"""Ensure file is completed."""

ensure_uncompleted = ensure_state(lambda o: not o.completed, MultipartAlreadyCompleted)
"""Ensure file is not completed."""

ensure_not_deleted = ensure_state(
    lambda o: not o.deleted,
    InvalidOperationError,
    [BucketError("Cannot make snapshot of a deleted bucket.")],
)
"""Ensure file is not deleted."""

ensure_unlocked = ensure_state(lambda o: not o.locked, BucketLockedError)
"""Ensure bucket is locked."""

ensure_no_file = ensure_state(lambda o: o.file_id is None, FileInstanceAlreadySetError)
"""Ensure file is not already set."""

ensure_is_previous_version = ensure_state(
    lambda o: not o.is_head,
    InvalidOperationError,
    [ObjectVersionError("Cannot restore latest version.")],
)
"""Ensure file is the previous version."""


#
# Model definitions
#
class Timestamp(object):
    """Timestamp model mix-in with fractional seconds support.

    SQLAlchemy-Utils timestamp model, does not have support for fractional
    seconds.
    """

    created = db.Column(
        db.DateTime().with_variant(mysql.DATETIME(fsp=6), "mysql"),
        default=datetime.utcnow,
        nullable=False,
    )
    """Creation timestamp."""

    updated = db.Column(
        db.DateTime().with_variant(mysql.DATETIME(fsp=6), "mysql"),
        default=datetime.utcnow,
        nullable=False,
    )
    """Modification timestamp."""


@db.event.listens_for(Timestamp, "before_update", propagate=True)
def timestamp_before_update(mapper, connection, target):
    """Listen for updating updated field."""
    target.updated = datetime.utcnow()


class Location(db.Model, Timestamp):
    """Model defining base locations."""

    __tablename__ = "files_location"

    id = db.Column(db.Integer, primary_key=True)
    """Internal identifier for locations.

    The internal identifier is used only used as foreign key for buckets in
    order to decrease storage requirements per row for buckets.
    """

    name = db.Column(db.String(20), unique=True, nullable=False)
    """External identifier of the location."""

    uri = db.Column(db.String(255), nullable=False)
    """URI of the location."""

    default = db.Column(db.Boolean(name="default"), nullable=False, default=False)
    """True if the location is the default location.

    At least one location should be the default location.
    """

    @validates("name")
    def validate_name(self, key, name):
        """Validate name."""
        if not slug_pattern.match(name) or len(name) > 20:
            raise ValueError(
                "Invalid location name (lower-case alphanumeric + dashes)."
            )
        return name

    @classmethod
    def get_by_name(cls, name):
        """Fetch a specific location object."""
        return cls.query.filter_by(
            name=name,
        ).one_or_none()

    @classmethod
    def get_default(cls):
        """Fetch the default location object."""
        try:
            return cls.query.filter_by(default=True).one_or_none()
        except MultipleResultsFound:
            return None

    @classmethod
    def all(cls):
        """Return query that fetches all locations."""
        return Location.query.all()

    def __repr__(self):
        """Return representation of location."""
        return self.name


class Bucket(db.Model, Timestamp):
    """Model for storing buckets.

    A bucket is a container of objects. Buckets have a default location and
    storage class. Individual objects in the bucket can however have different
    locations and storage classes.

    A bucket can be marked as deleted. A bucket can also be marked as locked
    to prevent operations on the bucket.

    Each bucket can also define a quota. The size of a bucket is the size
    of all objects in the bucket (including all versions).
    """

    __tablename__ = "files_bucket"

    id = db.Column(
        UUIDType,
        primary_key=True,
        default=uuid.uuid4,
    )
    """Bucket identifier."""

    default_location = db.Column(
        db.Integer, db.ForeignKey(Location.id, ondelete="RESTRICT"), nullable=False
    )
    """Default location."""

    default_storage_class = db.Column(
        db.String(1),
        nullable=False,
        default=lambda: current_app.config["FILES_REST_DEFAULT_STORAGE_CLASS"],
    )
    """Default storage class."""

    size = db.Column(db.BigInteger, default=0, nullable=False)
    """Size of bucket.

    This is a computed property which can rebuilt any time from the objects
    inside the bucket.
    """

    quota_size = db.Column(
        db.BigInteger,
        nullable=True,
        default=lambda: current_app.config["FILES_REST_DEFAULT_QUOTA_SIZE"],
    )
    """Quota size of bucket.

    Usage of this property depends on which file size limiters are installed.
    """

    max_file_size = db.Column(
        db.BigInteger,
        nullable=True,
        default=lambda: current_app.config["FILES_REST_DEFAULT_MAX_FILE_SIZE"],
    )
    """Maximum size of a single file in the bucket.

    Usage of this property depends on which file size limiters are installed.
    """

    locked = db.Column(db.Boolean(name="locked"), default=False, nullable=False)
    """Lock state of bucket.

    Modifications are not allowed on a locked bucket.
    """

    deleted = db.Column(db.Boolean(name="deleted"), default=False, nullable=False)
    """Delete state of bucket."""

    location = db.relationship(Location, backref="buckets")
    """Location associated with this bucket."""

    def __repr__(self):
        """Return representation of location."""
        return str(self.id)

    @property
    def quota_left(self):
        """Get how much space is left in the bucket."""
        if self.quota_size:
            return max(self.quota_size - self.size, 0)

    @property
    def size_limit(self):
        """Get size limit for this bucket.

        The limit is based on the minimum output of the file size limiters.
        """
        limits = [
            lim
            for lim in current_files_rest.file_size_limiters(self)
            if lim.limit is not None
        ]
        return min(limits) if limits else None

    @validates("default_storage_class")
    def validate_storage_class(self, key, default_storage_class):
        """Validate storage class."""
        if (
            default_storage_class
            not in current_app.config["FILES_REST_STORAGE_CLASS_LIST"]
        ):
            raise ValueError("Invalid storage class.")
        return default_storage_class

    @ensure_not_deleted()
    def snapshot(self, lock=False):
        """Create a snapshot of latest objects in bucket.

        :param lock: Create the new bucket in a locked state.
        :returns: Newly created bucket containing copied ObjectVersion.
        """
        with db.session.begin_nested():
            bucket = Bucket(
                default_location=self.default_location,
                default_storage_class=self.default_storage_class,
                quota_size=self.quota_size,
            )
            db.session.add(bucket)

        for o in ObjectVersion.get_by_bucket(self):
            o.copy(bucket=bucket)

        bucket.locked = True if lock else self.locked

        return bucket

    @ensure_not_deleted(msg=[BucketError("Cannot sync a deleted bucket.")])
    def sync(self, bucket, delete_extras=False):
        """Sync self bucket ObjectVersions to the destination bucket.

        The bucket is fully mirrored with the destination bucket following the
        logic:

        * same ObjectVersions are not touched
        * new ObjectVersions are added to destination
        * deleted ObjectVersions are deleted in destination
        * extra ObjectVersions in dest are deleted if `delete_extras` param is True

        :param bucket: The destination bucket.
        :param delete_extras: Delete extra ObjectVersions in destination if
            True.
        :returns: A tuple with the bucket having an exact copy of ObjectVersions in
            self, and a list with the changes applied in the format (action, value).
            Specifically, there are 2 possible actions:
            * "add": the`value` is the new copied object version.
            * "delete": the `value` is the key of the deleted object version.

        """
        assert not bucket.locked

        src_ovs = ObjectVersion.get_by_bucket(bucket=self, with_deleted=True)
        dest_ovs = ObjectVersion.get_by_bucket(bucket=bucket, with_deleted=True)
        changed_ovs = []

        # transform into a dict { key: object version }
        src_dict_ovs = {}
        for ov in src_ovs:
            # keep only the latest version of an ObjectVersion and don't
            # override with previous
            if ov.key not in src_dict_ovs:
                src_dict_ovs[ov.key] = ov

        dest_dict_ovs = {}
        for ov in dest_ovs:
            # keep only the latest version of an ObjectVersion and don't
            # override with previous
            if ov.key not in dest_dict_ovs:
                dest_dict_ovs[ov.key] = ov

        for key, ov in src_dict_ovs.items():
            key_in_dest = key in dest_dict_ovs
            if ov.deleted:
                if key_in_dest and not dest_dict_ovs[key].deleted:
                    ObjectVersion.delete(bucket, key)
                    changed_ovs.append(("delete", key))
            else:
                if not key_in_dest:
                    new_ov = ov.copy(bucket=bucket)
                    changed_ovs.append(("add", new_ov))
                else:
                    dest_ov = dest_dict_ovs[key]
                    file_in_dest_differs = ov.file_id != dest_ov.file_id
                    ov_deleted_in_dest = dest_ov.deleted
                    if file_in_dest_differs or ov_deleted_in_dest:
                        new_ov = ov.copy(bucket=bucket)
                        changed_ovs.append(("add", new_ov))

        if delete_extras:
            for key, ov in dest_dict_ovs.items():
                if key not in src_dict_ovs:
                    ObjectVersion.delete(bucket, key)
                    changed_ovs.append(("delete", key))

        return bucket, changed_ovs

    def get_tags(self):
        """Get tags for bucket as dictionary."""
        return {t.key: t.value for t in self.tags}

    @classmethod
    def create(cls, location=None, storage_class=None, **kwargs):
        r"""Create a bucket.

        :param location: Location of a bucket (instance or name).
            Default: Default location.
        :param storage_class: Storage class of a bucket.
            Default: Default storage class.
        :param \**kwargs: Keyword arguments are forwarded to the class
        :param \**kwargs: Keyword arguments are forwarded to the class
            constructor.
        :returns: Created bucket.
        """
        with db.session.begin_nested():
            if location is None:
                location = Location.get_default()
            elif isinstance(location, str):
                location = Location.get_by_name(location)

            obj = cls(
                default_location=location.id,
                default_storage_class=storage_class
                or current_app.config["FILES_REST_DEFAULT_STORAGE_CLASS"],
                **kwargs
            )
            db.session.add(obj)
        return obj

    @classmethod
    def get(cls, bucket_id):
        """Get a bucket object (excluding deleted).

        :param bucket_id: Bucket identifier.
        :returns: Bucket instance.
        """
        return cls.query.filter_by(id=bucket_id, deleted=False).one_or_none()

    @classmethod
    def all(cls):
        """Return query of all buckets (excluding deleted)."""
        return cls.query.filter_by(deleted=False)

    @classmethod
    def delete(cls, bucket_id):
        """Delete a bucket.

        Does not actually delete the Bucket, just marks it as deleted.
        """
        bucket = cls.get(bucket_id)
        if not bucket or bucket.deleted:
            return False

        bucket.deleted = True
        return True

    @ensure_unlocked()
    def remove(self):
        """Permanently remove a bucket and all objects (including versions).

        .. warning::

           This by-passes the normal versioning and should only be used when
           you want to permanently delete a bucket and its objects. Otherwise
           use :py:data:`Bucket.delete()`.

           Note the method does not remove the associated file instances which
           must be garbage collected.

        :returns: ``self``.
        """
        with db.session.begin_nested():
            ObjectVersion.query.filter_by(bucket_id=self.id).delete()
            self.query.filter_by(id=self.id).delete()
        return self


class BucketTag(db.Model):
    """Model for storing tags associated to buckets.

    This is useful to store extra information for a bucket.
    """

    __tablename__ = "files_buckettags"

    bucket_id = db.Column(
        UUIDType,
        db.ForeignKey(Bucket.id, ondelete="CASCADE"),
        default=uuid.uuid4,
        primary_key=True,
    )

    key = db.Column(db.String(255), primary_key=True)
    """Tag key."""

    value = db.Column(db.Text, nullable=False)
    """Tag value."""

    bucket = db.relationship(Bucket, backref="tags")
    """Relationship to buckets."""

    @classmethod
    def get(cls, bucket, key):
        """Get tag object."""
        return cls.query.filter_by(
            bucket_id=as_bucket_id(bucket),
            key=key,
        ).one_or_none()

    @classmethod
    def create(cls, bucket, key, value):
        """Create a new tag for bucket."""
        with db.session.begin_nested():
            obj = cls(bucket_id=as_bucket_id(bucket), key=key, value=value)
            db.session.add(obj)
        return obj

    @classmethod
    def create_or_update(cls, bucket, key, value):
        """Create or update a new tag for bucket."""
        obj = cls.get(bucket, key)
        if obj:
            obj.value = value
            db.session.merge(obj)
        else:
            obj = cls.create(bucket, key, value)
        return obj

    @classmethod
    def get_value(cls, bucket, key):
        """Get tag value."""
        obj = cls.get(bucket, key)
        return obj.value if obj else None

    @classmethod
    def delete(cls, bucket, key):
        """Delete a tag."""
        with db.session.begin_nested():
            cls.query.filter_by(
                bucket_id=as_bucket_id(bucket),
                key=key,
            ).delete()


class FileInstance(db.Model, Timestamp):
    """Model for storing files.

    A file instance represents a file on disk. A file instance may be linked
    from many objects, while an object can have one and only one file instance.

    A file instance also records the storage class, size and checksum of the
    file on disk.

    Additionally, a file instance can be read only in case the storage layer
    is not capable of writing to the file (e.g. can typically be used to
    link to files on externally controlled storage).
    """

    __tablename__ = "files_files"

    id = db.Column(
        UUIDType,
        primary_key=True,
        default=uuid.uuid4,
    )
    """Identifier of file."""

    uri = db.Column(
        db.Text().with_variant(mysql.VARCHAR(255), "mysql"), unique=True, nullable=True
    )
    """Location of file."""

    storage_class = db.Column(db.String(1), nullable=True)
    """Storage class of file."""

    size = db.Column(db.BigInteger, default=0, nullable=True)
    """Size of file."""

    checksum = db.Column(db.String(255), nullable=True)
    """String representing the checksum of the object."""

    readable = db.Column(db.Boolean(name="readable"), default=True, nullable=False)
    """Defines if the file is read only."""

    writable = db.Column(db.Boolean(name="writable"), default=True, nullable=False)
    """Defines if file is writable.

    This property is used to create a file instance prior to having the actual
    file at the given URI. This is useful when e.g. copying a file instance.
    """

    last_check_at = db.Column(db.DateTime, nullable=True)
    """Timestamp of last fixity check."""

    last_check = db.Column(db.Boolean(name="last_check"), default=True)
    """Result of last fixity check."""

    @validates("uri")
    def validate_uri(self, key, uri):
        """Validate uri."""
        if len(uri) > current_app.config["FILES_REST_FILE_URI_MAX_LEN"]:
            raise ValueError("FileInstance URI too long ({0}).".format(len(uri)))
        return uri

    @classmethod
    def get(cls, file_id):
        """Get a file instance."""
        return cls.query.filter_by(id=file_id).one_or_none()

    @classmethod
    def get_by_uri(cls, uri):
        """Get a file instance by URI."""
        assert uri is not None
        return cls.query.filter_by(uri=uri).one_or_none()

    @classmethod
    def create(cls):
        """Create a file instance.

        Note, object is only added to the database session.
        """
        obj = cls(
            id=uuid.uuid4(),
            writable=True,
            readable=False,
            size=0,
        )
        db.session.add(obj)
        return obj

    def delete(self):
        """Delete a file instance.

        The file instance can be deleted if it has no references from other
        objects. The caller is responsible to test if the file instance is
        writable and that the disk file can actually be removed.

        .. note::

           Normally you should use the Celery task to delete a file instance,
           as this method will not remove the file on disk.
        """
        self.query.filter_by(id=self.id).delete()
        return self

    def storage(self, **kwargs):
        """Get storage interface for object.

        Uses the applications storage factory to create a storage interface
        that can be used for this particular file instance.

        :returns: Storage interface.
        """
        return current_files_rest.storage_factory(fileinstance=self, **kwargs)

    @ensure_readable()
    def update_checksum(
        self, progress_callback=None, chunk_size=None, checksum_kwargs=None, **kwargs
    ):
        """Update checksum based on file."""
        self.checksum = self.storage(**kwargs).checksum(
            progress_callback=progress_callback,
            chunk_size=chunk_size,
            **(checksum_kwargs or {})
        )

    def clear_last_check(self):
        """Clear the checksum of the file."""
        with db.session.begin_nested():
            self.last_check = None
            self.last_check_at = datetime.utcnow()
        return self

    def verify_checksum(
        self,
        progress_callback=None,
        chunk_size=None,
        throws=True,
        checksum_kwargs=None,
        **kwargs
    ):
        """Verify checksum of file instance.

        :param bool throws: If `True`, exceptions raised during checksum
            calculation will be re-raised after logging. If set to `False`, and
            an exception occurs, the `last_check` field is set to `None`
            (`last_check_at` of course is updated), since no check actually was
            performed.
        :param dict checksum_kwargs: Passed as `**kwargs`` to
            ``storage().checksum``.
        """
        try:
            real_checksum = self.storage(**kwargs).checksum(
                progress_callback=progress_callback,
                chunk_size=chunk_size,
                **(checksum_kwargs or {})
            )
        except Exception as exc:
            current_app.logger.exception(str(exc))
            if throws:
                raise
            real_checksum = None
        with db.session.begin_nested():
            self.last_check = (
                None if real_checksum is None else (self.checksum == real_checksum)
            )
            self.last_check_at = datetime.utcnow()
        return self.last_check

    @ensure_writable()
    def init_contents(self, size=0, **kwargs):
        """Initialize file."""
        self.set_uri(
            *self.storage(**kwargs).initialize(size=size), readable=False, writable=True
        )

    @ensure_writable()
    def update_contents(
        self,
        stream,
        seek=0,
        size=None,
        chunk_size=None,
        progress_callback=None,
        **kwargs
    ):
        """Save contents of stream to this file.

        :param obj: ObjectVersion instance from where this file is accessed
            from.
        :param stream: File-like stream.
        """
        self.checksum = None
        return self.storage(**kwargs).update(
            stream,
            seek=seek,
            size=size,
            chunk_size=chunk_size,
            progress_callback=progress_callback,
        )

    @ensure_writable()
    def set_contents(
        self,
        stream,
        chunk_size=None,
        size=None,
        size_limit=None,
        progress_callback=None,
        **kwargs
    ):
        """Save contents of stream to this file.

        :param obj: ObjectVersion instance from where this file is accessed
            from.
        :param stream: File-like stream.
        """
        self.set_uri(
            *self.storage(**kwargs).save(
                stream,
                chunk_size=chunk_size,
                size=size,
                size_limit=size_limit,
                progress_callback=progress_callback,
            )
        )

    @ensure_writable()
    def copy_contents(
        self, fileinstance, progress_callback=None, chunk_size=None, **kwargs
    ):
        """Copy this file instance into another file instance."""
        if not fileinstance.readable:
            raise ValueError("Source file instance is not readable.")
        if not self.size == 0:
            raise ValueError("File instance has data.")

        self.set_uri(
            *self.storage(**kwargs).copy(
                fileinstance.storage(**kwargs),
                chunk_size=chunk_size,
                progress_callback=progress_callback,
            )
        )

    @ensure_readable()
    def send_file(
        self,
        filename,
        restricted=True,
        mimetype=None,
        trusted=False,
        chunk_size=None,
        as_attachment=False,
        **kwargs
    ):
        """Send file to client."""
        return self.storage(**kwargs).send_file(
            filename,
            mimetype=mimetype,
            restricted=restricted,
            checksum=self.checksum,
            trusted=trusted,
            chunk_size=chunk_size,
            as_attachment=as_attachment,
        )

    def set_uri(
        self, uri, size, checksum, readable=True, writable=False, storage_class=None
    ):
        """Set a location of a file."""
        self.uri = uri
        self.size = size
        self.checksum = checksum
        self.writable = writable
        self.readable = readable
        self.storage_class = (
            current_app.config["FILES_REST_DEFAULT_STORAGE_CLASS"]
            if storage_class is None
            else storage_class
        )
        return self


class ObjectVersion(db.Model, Timestamp):
    """Model for storing versions of objects.

    A bucket stores one or more objects identified by a key. Each object is
    versioned where each version is represented by an ``ObjectVersion``.

    An object version can either be 1) a *normal version* which is linked to
    a file instance, or 2) a *delete marker*, which is *not* linked to a file
    instance.

    An normal object version is linked to a physical file on disk via a file
    instance. This allows for multiple object versions to point to the same
    file on disk, to optimize storage efficiency (e.g. useful for snapshotting
    an entire bucket without duplicating the files).

    A delete marker object version represents that the object at hand was
    deleted.

    The latest version of an object is marked using the ``is_head`` property.
    If the latest object version is a delete marker the object will not be
    shown in the bucket.
    """

    __tablename__ = "files_object"

    version_id = db.Column(UUIDType, primary_key=True, default=uuid.uuid4)
    """Identifier for the specific version of an object."""

    key = db.Column(db.Text().with_variant(mysql.VARCHAR(255), "mysql"), nullable=False)
    """Key identifying the object."""

    bucket_id = db.Column(
        UUIDType,
        db.ForeignKey(Bucket.id, ondelete="RESTRICT"),
        default=uuid.uuid4,
        nullable=False,
    )
    """Bucket identifier."""

    file_id = db.Column(
        UUIDType, db.ForeignKey(FileInstance.id, ondelete="RESTRICT"), nullable=True
    )
    """File instance for this object version.

    A null value in this column defines that the object has been deleted.
    """

    _mimetype = db.Column(
        db.String(255),
        index=True,
        nullable=True,
    )
    """MIME type of the object."""

    is_head = db.Column(db.Boolean(name="is_head"), nullable=False, default=True)
    """Defines if object is the latest version."""

    # Relationships definitions
    bucket = db.relationship(Bucket, backref="objects")
    """Relationship to buckets."""

    file = db.relationship(FileInstance, backref="objects")
    """Relationship to file instance."""

    __table_args__ = (db.UniqueConstraint("bucket_id", "version_id", "key"),)

    @validates("key")
    def validate_key(self, key, key_):
        """Validate key."""
        return validate_key(key_)

    def __unicode__(self):
        """Return unicoded object."""
        return "{0}:{1}:{2}".format(self.bucket_id, self.version_id, self.key)

    # https://docs.python.org/3.3/howto/pyporting.html#str-unicode
    if sys.version_info[0] >= 3:  # Python 3

        def __repr__(self):
            """Return representation of location."""
            return self.__unicode__()

    else:  # Python 2

        def __repr__(self):
            """Return representation of location."""
            return self.__unicode__().encode("utf8")

    @hybrid_property
    def mimetype(self):
        """Get MIME type of object."""
        return self._mimetype if self._mimetype else guess_mimetype(self.key)

    @mimetype.setter
    def mimetype(self, value):
        """Setter for MIME type."""
        self._mimetype = value

    @property
    def basename(self):
        """Return filename of the object."""
        return basename(self.key)

    @property
    def deleted(self):
        """Determine if object version is a delete marker."""
        return self.file_id is None

    @ensure_no_file()
    @update_bucket_size
    def set_contents(
        self,
        stream,
        chunk_size=None,
        size=None,
        size_limit=None,
        progress_callback=None,
    ):
        """Save contents of stream to file instance.

        If a file instance has already been set, this methods raises an
        ``FileInstanceAlreadySetError`` exception.

        :param stream: File-like stream.
        :param size: Size of stream if known.
        :param chunk_size: Desired chunk size to read stream in. It is up to
            the storage interface if it respects this value.
        """
        if size_limit is None:
            size_limit = self.bucket.size_limit

        self.file = FileInstance.create()
        self.file.set_contents(
            stream,
            size_limit=size_limit,
            size=size,
            chunk_size=chunk_size,
            progress_callback=progress_callback,
            default_location=self.bucket.location.uri,
            default_storage_class=self.bucket.default_storage_class,
        )

        return self

    @ensure_no_file()
    @update_bucket_size
    def set_location(self, uri, size, checksum, storage_class=None):
        """Set only URI location of for object.

        Useful to link files on externally controlled storage. If a file
        instance has already been set, this methods raises an
        ``FileInstanceAlreadySetError`` exception.

        :param uri: Full URI to object (which can be interpreted by the storage
            interface).
        :param size: Size of file.
        :param checksum: Checksum of file.
        :param storage_class: Storage class where file is stored ()
        """
        self.file = FileInstance()
        self.file.set_uri(uri, size, checksum, storage_class=storage_class)
        db.session.add(self.file)
        return self

    @ensure_no_file()
    @update_bucket_size
    def set_file(self, fileinstance):
        """Set a file instance."""
        self.file = fileinstance
        return self

    def send_file(self, restricted=True, trusted=False, **kwargs):
        """Wrap around FileInstance's send file."""
        return self.file.send_file(
            self.basename,
            restricted=restricted,
            mimetype=self.mimetype,
            trusted=trusted,
            **kwargs
        )

    @ensure_is_previous_version()
    def restore(self):
        """Restore this object version to become the latest version.

        Raises an exception if the object is the latest version.
        """
        # Note, copy calls create which will fail if bucket is locked.
        return self.copy()

    @ensure_not_deleted(msg=[ObjectVersionError("Cannot copy a delete marker.")])
    def copy(self, bucket=None, key=None):
        """Copy an object version to a given bucket + object key.

        The copy operation is handled completely at the metadata level. The
        actual data on disk is not copied. Instead, the two object versions
        will point to the same physical file (via the same FileInstance).

        All the tags associated with the current object version are copied over
        to the new instance.

        .. warning::

           If the destination object exists, it will be replaced by the new
           object version which will become the latest version.

        :param bucket: The bucket (instance or id) to copy the object to.
            Default: current bucket.
        :param key: Key name of destination object.
            Default: current object key.
        :returns: The copied object version.
        """
        new_ob = ObjectVersion.create(
            self.bucket if bucket is None else as_bucket(bucket),
            key or self.key,
            _file_id=self.file_id,
        )

        for tag in self.tags:
            ObjectVersionTag.create_or_update(
                object_version=new_ob, key=tag.key, value=tag.value
            )

        return new_ob

    @ensure_unlocked(getter=lambda o: not o.bucket.locked)
    def remove(self):
        """Permanently remove a specific object version from the database.

        .. warning::

           This by-passes the normal versioning and should only be used when
           you want to permanently delete a specific object version. Otherwise
           use :py:data:`ObjectVersion.delete()`.

           Note the method does not remove the associated file instance which
           must be garbage collected.

        :returns: ``self``.
        """
        with db.session.begin_nested():
            if self.file_id:
                self.bucket.size -= self.file.size
            self.query.filter_by(
                bucket_id=self.bucket_id,
                key=self.key,
                version_id=self.version_id,
            ).delete()

        return self

    @classmethod
    def create(
        cls,
        bucket,
        key,
        _file_id=None,
        stream=None,
        mimetype=None,
        version_id=None,
        **kwargs
    ):
        """Create a new object in a bucket.

        The created object is by default created as a delete marker. You must
        use ``set_contents()`` or ``set_location()`` in order to change this.

        :param bucket: The bucket (instance or id) to create the object in.
        :param key: Key of object.
        :param _file_id: For internal use.
        :param stream: File-like stream object. Used to set content of object
            immediately after being created.
        :param mimetype: MIME type of the file object if it is known.
        :param kwargs: Keyword arguments passed to ``Object.set_contents()``.
        """
        bucket = as_bucket(bucket)

        if bucket.locked:
            raise BucketLockedError()

        with db.session.begin_nested():
            latest_obj = cls.query.filter(
                cls.bucket == bucket, cls.key == key, cls.is_head.is_(True)
            ).one_or_none()
            if latest_obj is not None:
                latest_obj.is_head = False
                db.session.add(latest_obj)

            # By default objects are created in a deleted state (i.e.
            # file_id is null).
            obj = cls(
                bucket=bucket,
                key=key,
                version_id=version_id or uuid.uuid4(),
                is_head=True,
                mimetype=mimetype,
            )
            if _file_id:
                file_ = (
                    _file_id
                    if isinstance(_file_id, FileInstance)
                    else FileInstance.get(_file_id)
                )
                obj.set_file(file_)
            db.session.add(obj)
        if stream:
            obj.set_contents(stream, **kwargs)
        return obj

    @classmethod
    def get(cls, bucket, key, version_id=None):
        """Fetch a specific object.

        By default the latest object version is returned, if
        ``version_id`` is not set.

        :param bucket: The bucket (instance or id) to get the object from.
        :param key: Key of object.
        :param version_id: Specific version of an object.
        """
        filters = [
            cls.bucket_id == as_bucket_id(bucket),
            cls.key == key,
        ]

        if version_id:
            filters.append(cls.version_id == version_id)
        else:
            filters.append(cls.is_head.is_(True))
            filters.append(cls.file_id.isnot(None))

        return cls.query.filter(*filters).one_or_none()

    @classmethod
    def get_versions(cls, bucket, key, desc=True):
        """Fetch all versions of a specific object.

        :param bucket: The bucket (instance or id) to get the object from.
        :param key: Key of object.
        :param desc: Sort results desc if True, asc otherwise.
        :returns: The query to execute to fetch all versions.
        """
        filters = [
            cls.bucket_id == as_bucket_id(bucket),
            cls.key == key,
        ]

        order = cls.created.desc() if desc else cls.created.asc()

        return cls.query.filter(*filters).order_by(cls.key, order)

    @classmethod
    def delete(cls, bucket, key):
        """Delete an object.

        Technically works by creating a new version which works as a delete
        marker.

        :param bucket: The bucket (instance or id) to delete the object from.
        :param key: Key of object.
        :returns: Created delete marker object if key exists else ``None``.
        """
        bucket = as_bucket(bucket)

        obj = cls.get(bucket.id, key)
        if obj:
            bucket.size -= obj.file.size
            return cls.create(bucket, key)
        return None

    @classmethod
    def get_by_bucket(cls, bucket, versions=False, with_deleted=False):
        """Return query that fetches all the objects in a bucket.

        :param bucket: The bucket (instance or id) to query.
        :param versions: Select all versions if True, only heads otherwise.
        :param with_deleted: Select also deleted objects if True.
        :returns: The query to retrieve filtered objects in the given bucket.
        """
        bucket_id = bucket.id if isinstance(bucket, Bucket) else bucket

        filters = [
            cls.bucket_id == bucket_id,
        ]

        if not versions:
            filters.append(cls.is_head.is_(True))

        if not with_deleted:
            filters.append(cls.file_id.isnot(None))

        return cls.query.filter(*filters).order_by(cls.key, cls.created.desc())

    @classmethod
    def relink_all(cls, old_file, new_file):
        """Relink all object versions (for a given file) to a new file.

        .. warning::

           Use this method with great care.
        """
        assert old_file.checksum == new_file.checksum
        assert old_file.id
        assert new_file.id

        with db.session.begin_nested():
            ObjectVersion.query.filter_by(file_id=str(old_file.id)).update(
                {ObjectVersion.file_id: str(new_file.id)}
            )

    def get_tags(self):
        """Get tags for object version as dictionary."""
        return {t.key: t.value for t in self.tags}

    def __eq__(self, other):
        """Check if the two object are equals."""
        return (
            other
            and isinstance(other, self.__class__)
            and self.key == other.key
            and self.file_id == other.file_id
        )

    def __ne__(self, other):
        """Check if are not equal."""
        return not self.__eq__(other=other)

    # DDL string used to avoid automap in mysql until sqlalchemy 2.0
    # https://github.com/sqlalchemy/sqlalchemy/discussions/7597
    @classmethod
    def ix_uq_partial_files_object_is_head_dll(cls):
        """Return DDL instruction for ix_uq_partial_files_object_is_head."""
        return db.DDL(
            "CREATE UNIQUE INDEX ix_uq_partial_files_object_is_head "
            "ON %(table)s (bucket_id, key) WHERE is_head"
        )


db.event.listen(
    ObjectVersion.__table__,
    "after_create",
    ObjectVersion.ix_uq_partial_files_object_is_head_dll().execute_if(
        dialect="postgresql"
    ),
)
"""Create ix_uq_partial_files_object_is_head only on postgresql backend."""


class ObjectVersionTag(db.Model):
    """Model for storing tags associated to object versions.

    Used for storing extra technical information for an object version.
    """

    __tablename__ = "files_objecttags"

    version_id = db.Column(
        UUIDType,
        db.ForeignKey(ObjectVersion.version_id, ondelete="CASCADE"),
        default=uuid.uuid4,
        primary_key=True,
    )
    """Object version id."""

    key = db.Column(db.String(255), primary_key=True)
    """Tag key."""

    value = db.Column(db.Text, nullable=False)
    """Tag value."""

    object_version = db.relationship(ObjectVersion, backref="tags")
    """Relationship to object versions."""

    def copy(self, object_version=None, key=None):
        """Copy a tag to a given object version.

        :param object_version: The object version instance to copy the tag to.
            Default: current object version.
        :param key: Key of destination tag.
            Default: current tag key.
        :return: The copied object version tag.
        """
        return ObjectVersionTag.create(
            self.object_version if object_version is None else object_version,
            key or self.key,
            self.value,
        )

    @classmethod
    def get(cls, object_version, key):
        """Get the tag object."""
        return cls.query.filter_by(
            version_id=as_object_version_id(object_version),
            key=key,
        ).one_or_none()

    @classmethod
    def create(cls, object_version, key, value):
        """Create a new tag for a given object version."""
        assert len(key) < 256
        assert len(value) < 256
        with db.session.begin_nested():
            obj = cls(
                version_id=as_object_version_id(object_version), key=key, value=value
            )
            db.session.add(obj)
        return obj

    @classmethod
    def create_or_update(cls, object_version, key, value):
        """Create or update a new tag for a given object version."""
        assert len(key) < 256
        assert len(value) < 256
        obj = cls.get(object_version, key)
        if obj:
            obj.value = value
            db.session.merge(obj)
        else:
            obj = cls.create(object_version, key, value)
        return obj

    @classmethod
    def get_value(cls, object_version, key):
        """Get the tag value."""
        obj = cls.get(object_version, key)
        return obj.value if obj else None

    @classmethod
    def delete(cls, object_version, key=None):
        """Delete tags.

        :param object_version: The object version instance or id.
        :param key: Key of the tag to delete.
            Default: delete all tags.
        """
        with db.session.begin_nested():
            q = cls.query.filter_by(version_id=as_object_version_id(object_version))
            if key:
                q = q.filter_by(key=key)
            q.delete()


class MultipartObject(db.Model, Timestamp):
    """Model for storing files in chunks.

    A multipart object belongs to a specific bucket and key and is identified
    by an upload id. You can have multiple multipart uploads for the same
    bucket and key. Once all parts of a multipart object is uploaded, the state
    is changed to ``completed``. Afterwards it is not possible to upload
    new parts. Once completed, the multipart object is merged, and added as
    a new version in the current object/bucket.

    All parts for a multipart upload must be of the same size, except for the
    last part.
    """

    __tablename__ = "files_multipartobject"

    __table_args__ = (
        db.UniqueConstraint("upload_id", "bucket_id", "key", name="uix_item"),
    )

    upload_id = db.Column(
        UUIDType,
        default=uuid.uuid4,
        primary_key=True,
    )
    """Identifier for the specific version of an object."""

    bucket_id = db.Column(
        UUIDType,
        db.ForeignKey(Bucket.id, ondelete="RESTRICT"),
    )
    """Bucket identifier."""

    key = db.Column(
        db.Text().with_variant(mysql.VARCHAR(255), "mysql"),
    )
    """Key identifying the object."""

    file_id = db.Column(
        UUIDType, db.ForeignKey(FileInstance.id, ondelete="RESTRICT"), nullable=False
    )
    """File instance for this multipart object."""

    chunk_size = db.Column(db.Integer, nullable=True)
    """Size of chunks for file."""

    size = db.Column(db.BigInteger, nullable=True)
    """Size of file."""

    completed = db.Column(db.Boolean(name="completed"), nullable=False, default=False)
    """Defines if object is the completed."""

    # Relationships definitions
    bucket = db.relationship(Bucket, backref="multipart_objects")
    """Relationship to buckets."""

    file = db.relationship(FileInstance, backref="multipart_objects")
    """Relationship to buckets."""

    def __repr__(self):
        """Return representation of the multipart object."""
        return "{0}:{2}:{1}".format(self.bucket_id, self.key, self.upload_id)

    @property
    def last_part_number(self):
        """Get last part number."""
        return (
            int(self.size / self.chunk_size)
            if self.size % self.chunk_size
            else int(self.size / self.chunk_size) - 1
        )

    @property
    def last_part_size(self):
        """Get size of last part."""
        return self.size % self.chunk_size

    @validates("key")
    def validate_key(self, key, key_):
        """Validate key."""
        return validate_key(key_)

    @staticmethod
    def is_valid_chunksize(chunk_size):
        """Check if size is valid."""
        min_csize = current_app.config["FILES_REST_MULTIPART_CHUNKSIZE_MIN"]
        max_csize = current_app.config["FILES_REST_MULTIPART_CHUNKSIZE_MAX"]
        return chunk_size >= min_csize and chunk_size <= max_csize

    @staticmethod
    def is_valid_size(size, chunk_size):
        """Validate max theoretical size."""
        min_csize = current_app.config["FILES_REST_MULTIPART_CHUNKSIZE_MIN"]
        max_size = chunk_size * current_app.config["FILES_REST_MULTIPART_MAX_PARTS"]
        return size > min_csize and size <= max_size

    def expected_part_size(self, part_number):
        """Get expected part size for a particular part number."""
        last_part = self.multipart.last_part_number

        if part_number == last_part:
            return self.multipart.last_part_size
        elif part_number >= 0 and part_number < last_part:
            return self.multipart.chunk_size
        else:
            raise MultipartInvalidPartNumber()

    @ensure_uncompleted()
    def complete(self):
        """Mark a multipart object as complete."""
        if Part.count(self) != self.last_part_number + 1:
            raise MultipartMissingParts()

        with db.session.begin_nested():
            self.completed = True
            self.file.readable = True
            self.file.writable = False
        return self

    @ensure_completed()
    def merge_parts(self, version_id=None, **kwargs):
        """Merge parts into object version."""
        self.file.update_checksum(**kwargs)
        with db.session.begin_nested():
            obj = ObjectVersion.create(
                self.bucket, self.key, _file_id=self.file_id, version_id=version_id
            )
            self.delete()
        return obj

    def delete(self):
        """Delete a multipart object."""
        # Update bucket size.
        self.bucket.size -= self.size
        # Remove parts
        Part.query_by_multipart(self).delete()
        # Remove self
        self.query.filter_by(upload_id=self.upload_id).delete()

    @classmethod
    def create(cls, bucket, key, size, chunk_size):
        """Create a new object in a bucket."""
        bucket = as_bucket(bucket)

        if bucket.locked:
            raise BucketLockedError()

        # Validate chunk size.
        if not cls.is_valid_chunksize(chunk_size):
            raise MultipartInvalidChunkSize()

        # Validate max theoretical size.
        if not cls.is_valid_size(size, chunk_size):
            raise MultipartInvalidSize()

        # Validate max bucket size.
        bucket_limit = bucket.size_limit
        if bucket_limit and size > bucket_limit:
            desc = (
                "File size limit exceeded."
                if isinstance(bucket_limit, int)
                else bucket_limit.reason
            )
            raise FileSizeError(description=desc)

        with db.session.begin_nested():
            file_ = FileInstance.create()
            file_.size = size
            obj = cls(
                upload_id=uuid.uuid4(),
                bucket=bucket,
                key=key,
                chunk_size=chunk_size,
                size=size,
                completed=False,
                file=file_,
            )
            bucket.size += size
            db.session.add(obj)
        file_.init_contents(
            size=size,
            default_location=bucket.location.uri,
            default_storage_class=bucket.default_storage_class,
        )
        return obj

    @classmethod
    def get(cls, bucket, key, upload_id, with_completed=False):
        """Fetch a specific multipart object."""
        q = cls.query.filter_by(
            upload_id=upload_id,
            bucket_id=as_bucket_id(bucket),
            key=key,
        )
        if not with_completed:
            q = q.filter(cls.completed.is_(False))

        return q.one_or_none()

    @classmethod
    def query_expired(cls, dt, bucket=None):
        """Query all uncompleted multipart uploads."""
        q = cls.query.filter(cls.created < dt).filter_by(completed=True)
        if bucket:
            q = q.filter(cls.bucket_id == as_bucket_id(bucket))
        return q

    @classmethod
    def query_by_bucket(cls, bucket):
        """Query all uncompleted multipart uploads."""
        return cls.query.filter(cls.bucket_id == as_bucket_id(bucket))


class Part(db.Model, Timestamp):
    """Part object."""

    __tablename__ = "files_multipartobject_part"

    upload_id = db.Column(
        UUIDType,
        db.ForeignKey(MultipartObject.upload_id, ondelete="RESTRICT"),
        primary_key=True,
    )
    """Multipart object identifier."""

    part_number = db.Column(db.Integer, primary_key=True, autoincrement=False)
    """Part number."""

    checksum = db.Column(db.String(255), nullable=True)
    """String representing the checksum of the part."""

    # Relationships definitions
    multipart = db.relationship(MultipartObject, backref="parts")
    """Relationship to multipart objects."""

    @property
    def start_byte(self):
        """Get start byte in file of this part."""
        return self.part_number * self.multipart.chunk_size

    @property
    def end_byte(self):
        """Get end byte in file for this part."""
        return min(
            (self.part_number + 1) * self.multipart.chunk_size, self.multipart.size
        )

    @property
    def part_size(self):
        """Get size of this part."""
        return self.end_byte - self.start_byte

    @classmethod
    def create(cls, mp, part_number, stream=None, **kwargs):
        """Create a new part object in a multipart object."""
        if part_number < 0 or part_number > mp.last_part_number:
            raise MultipartInvalidPartNumber()

        with db.session.begin_nested():
            obj = cls(
                multipart=mp,
                part_number=part_number,
            )
            db.session.add(obj)
        if stream:
            obj.set_contents(stream, **kwargs)
        return obj

    @classmethod
    def get_or_none(cls, mp, part_number):
        """Get part number."""
        return cls.query.filter_by(
            upload_id=mp.upload_id, part_number=part_number
        ).one_or_none()

    @classmethod
    def get_or_create(cls, mp, part_number):
        """Get or create a part."""
        obj = cls.get_or_none(mp, part_number)
        if obj:
            return obj
        return cls.create(mp, part_number)

    @classmethod
    def delete(cls, mp, part_number):
        """Get part number."""
        return cls.query.filter_by(
            upload_id=mp.upload_id, part_number=part_number
        ).delete()

    @classmethod
    def query_by_multipart(cls, multipart):
        """Get all parts for a specific multipart upload.

        :param multipart: A :class:`invenio_files_rest.models.MultipartObject`
            instance.
        :returns: A :class:`invenio_files_rest.models.Part` instance.
        """
        upload_id = (
            multipart.upload_id if isinstance(multipart, MultipartObject) else multipart
        )
        return cls.query.filter_by(upload_id=upload_id)

    @classmethod
    def count(cls, mp):
        """Count number of parts for a given multipart object."""
        return cls.query_by_multipart(mp).count()

    @ensure_uncompleted(getter=lambda o: not o.multipart.completed)
    def set_contents(self, stream, progress_callback=None):
        """Save contents of stream to part of file instance.

        If a the MultipartObject is completed this methods raises an
        ``MultipartAlreadyCompleted`` exception.

        :param stream: File-like stream.
        :param size: Size of stream if known.
        :param chunk_size: Desired chunk size to read stream in. It is up to
            the storage interface if it respects this value.
        """
        size, checksum = self.multipart.file.update_contents(
            stream,
            seek=self.start_byte,
            size=self.part_size,
            progress_callback=progress_callback,
        )
        self.checksum = checksum
        return self


__all__ = (
    "Bucket",
    "BucketTag",
    "FileInstance",
    "Location",
    "MultipartObject",
    "ObjectVersion",
    "ObjectVersionTag",
    "Part",
)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2020 CERN.
# Copyright (C) 2020 University of Münster.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Click command-line interface for file management."""

import click
from click_default_group import DefaultGroup
from flask.cli import with_appcontext
from invenio_db import db

from .models import Location

#
# File management commands
#


@click.group()
def files():
    """File management commands."""


@files.group()
def bucket():
    """Manage buckets."""


@bucket.command()
@with_appcontext
def touch():
    """Create new bucket."""
    from .models import Bucket

    bucket = Bucket.create()
    db.session.commit()
    click.secho(str(bucket), fg="green")


@bucket.command()
@click.argument("source", type=click.Path(exists=True, resolve_path=True))
@click.argument("bucket")
@click.option("--checksum/--no-checksum", default=False)
@click.option("--key-prefix", default="")
@with_appcontext
def cp(source, bucket, checksum, key_prefix):
    """Create new bucket from all files in directory."""
    from .helpers import populate_from_path
    from .models import Bucket

    for object_version in populate_from_path(
        Bucket.get(bucket), source, checksum=checksum, key_prefix=key_prefix
    ):
        click.secho(str(object_version))
    db.session.commit()


def _unset_default_location():
    """Unset default location, if there is one."""
    default_location = Location.get_default()
    if default_location is not None:
        default_location.default = False


@files.group(cls=DefaultGroup, default="create")
def location():
    """Manage locations."""


@location.command()
@click.argument("name")
@click.argument("uri")
@click.option("--default", is_flag=True, default=False)
@with_appcontext
def create(name, uri, default):
    """Create a new location."""
    location = Location.get_by_name(name)
    if location:
        click.secho(
            "Location {} (uri: {} default: {}) already exists".format(
                location.name, location.uri, str(location.default)
            ),
            fg="yellow",
        )

    else:
        if default:
            _unset_default_location()

        location = Location(name=name, uri=uri, default=default)
        db.session.add(location)
        db.session.commit()
        click.secho(
            "Location {} {} as default {} stored in database".format(
                location.name, location.uri, str(location.default)
            ),
            fg="green",
        )


@location.command()
@with_appcontext
def list():
    """Return the list of locations."""
    locations = Location.all()
    for location in locations:
        click.secho(
            "{} {} as default {}".format(
                location.name, location.uri, str(location.default)
            ),
            fg="green",
        )


@location.command()
@click.argument("name")
@with_appcontext
def set_default(name):
    """Set default a location as default. The location must already exist.

    If another location was marked as default it unsets it.
    """
    location = Location.get_by_name(name)
    if location is not None:
        _unset_default_location()
        location.default = True
        db.session.commit()
        click.secho(
            "Location {} {} set as default ({})".format(
                location.name, location.uri, str(location.default)
            ),
            fg="green",
        )

    click.secho("Location {} not found".format(name), fg="red")



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2019 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test limiters."""

import pytest

from invenio_files_rest.limiters import FileSizeLimit


def test_file_size_limit_comparisons():
    """Test FileSizeLimit comparison operators."""
    bigger = FileSizeLimit(100, "big limit")
    smaller = FileSizeLimit(50, "small limit")

    assert bigger > smaller
    assert smaller < bigger
    assert bigger == bigger
    assert bigger == 100
    assert bigger > 50
    assert bigger < 150

    with pytest.raises(NotImplementedError):
        bigger > 90.25
    with pytest.raises(NotImplementedError):
        bigger < 90.25
    with pytest.raises(NotImplementedError):
        bigger == 90.25



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2019 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Files download/upload REST API similar to S3 for Invenio."""

from io import BytesIO


def login_user(client, user):
    """Log in a specified user."""
    with client.session_transaction() as sess:
        sess["_user_id"] = user.id if user else None
        sess["_fresh"] = True


class BadBytesIO(BytesIO):
    """Class for closing the stream for further reading abruptly."""

    def __init__(self, *args, **kwargs):
        """Initialize."""
        self.called = False
        return super(BadBytesIO, self).__init__(*args, **kwargs)

    def read(self, *args, **kwargs):
        """Fail on second read."""
        if self.called:
            self.close()
        self.called = True
        return super(BadBytesIO, self).read(*args, **kwargs)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2019 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test serializer."""

from marshmallow import fields

from invenio_files_rest.serializer import (
    BaseSchema,
    json_serializer,
    serializer_mapping,
)


def test_serialize_pretty(app):
    """Test pretty JSON."""

    class TestSchema(BaseSchema):
        title = fields.Str(attribute="title")

    data = {"title": "test"}
    context = {
        "bucket": "11111111-1111-1111-1111-111111111111",
        "class": "TestSchema",
        "many": False,
    }

    serializer_mapping["TestSchema"] = TestSchema

    # TODO This test should be checked if it shouldn't have
    #  BaseSchema instead of Schema
    with app.test_request_context():
        assert json_serializer(data=data, context=context).data == b'{"title":"test"}'

    with app.test_request_context("/?prettyprint=1"):
        assert (
            json_serializer(data=data, context=context).data
            == b'{\n  "title": "test"\n}'
        )



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2019 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio Files Rest module configuration file."""

from datetime import timedelta

from invenio_files_rest.helpers import create_file_streaming_redirect_response

MAX_CONTENT_LENGTH = 16 * 1024 * 1024
"""Maximum allowed content length for form data.

This value limits the maximum file upload size via multipart-formdata and is
a Flask configuration variable that by default is unlimited.  The value must
be larger than the maximum part size you want to accept via
application/multipart-formdata (used by e.g. ng-file upload). This value only
limits file upload size via application/multipart-formdata and in particular
does not restrict the maximum file size possible when streaming a file in the
body of a PUT request.

Flask, by default, saves any file bigger than 500kb to a temporary file on
disk, thus do not set this value to large or you may run out of disk space on
your nodes.
"""

FILES_REST_STORAGE_CLASS_LIST = {
    "S": "Standard",
    "A": "Archive",
}
"""Storage class list defines the systems storage classes.

Storage classes are useful for e.g. defining the type of storage an object
is located on (e.g. offline/online), so that the system known if it can serve
the file and/or what is the reliability.
"""

FILES_REST_DEFAULT_STORAGE_CLASS = "S"
"""Default storage class. Must be one of `FILES_REST_STORAGE_CLASS_LIST`."""

FILES_REST_DEFAULT_QUOTA_SIZE = None
"""Default quota size for a bucket in bytes. `None` if unlimited."""

FILES_REST_DEFAULT_MAX_FILE_SIZE = None
"""Default maximum file size for a bucket in bytes. `None` if unlimited."""

FILES_REST_MIN_FILE_SIZE = 1
"""Minimum file size when uploading, in bytes (do not allow empty files)."""

FILES_REST_SIZE_LIMITERS = "invenio_files_rest.limiters.file_size_limiters"
"""Import path of file size limiters factory to control bucket size limits."""

FILES_REST_STORAGE_FACTORY = "invenio_files_rest.storage.pyfs_storage_factory"
"""Import path of factory used to create a storage instance."""

FILES_REST_PERMISSION_FACTORY = "invenio_files_rest.permissions.permission_factory"
"""Permission factory to control the files access from the REST interface."""

FILES_REST_OBJECT_KEY_MAX_LEN = 255
"""Maximum length of the ObjectVersion.key field.

.. warning::
   Setting this variable to anything higher than 255 is only supported
   with PostgreSQL database.
"""

FILES_REST_FILE_URI_MAX_LEN = 255
"""Maximum length of the FileInstance.uri field.

.. warning::
   Setting this variable to anything higher than 255 is only supported
   with PostgreSQL database.
"""

FILES_REST_STORAGE_PATH_SPLIT_LENGTH = 2
"""Number of chars to use as folder name when generating the path of a file.

   For example, if split length set to 4 and dimension to 4, the final
   path will be `a2ad/4kc9/8j39-34jn/`.
"""

FILES_REST_STORAGE_PATH_DIMENSIONS = 2
"""Number of directory levels created when generating the path of a file.

   For example, if split length set to 2 and dimension to 3, the final
   path will be `a2/ad/4k/c9-8j39-34jn/`.
"""

FILES_REST_MULTIPART_PART_FACTORIES = [
    "invenio_files_rest.views:default_partfactory",
    "invenio_files_rest.views:ngfileupload_partfactory",
]
"""Import path of factories used when parsing upload params for multipart."""

FILES_REST_UPLOAD_FACTORIES = [
    "invenio_files_rest.views:stream_uploadfactory",
    "invenio_files_rest.views:ngfileupload_uploadfactory",
]
"""Import path of factories used when parsing upload parameters.

.. note::

   Factories that reads ``request.stream`` directly must be first in the list,
   otherwise Werkzeug's form-data parser will read the stream.
"""

FILES_REST_MULTIPART_MAX_PARTS = 10000
"""Maximum number of parts when uploading files with multipart uploads."""

FILES_REST_MULTIPART_CHUNKSIZE_MIN = 5 * 1024 * 1024  # 5 MiB
"""Minimum chunk size in bytes of multipart objects."""

FILES_REST_MULTIPART_CHUNKSIZE_MAX = 5 * 1024 * 1024 * 1024  # 5 GiB
"""Maximum chunk size in bytes of multipart objects."""

FILES_REST_MULTIPART_EXPIRES = timedelta(days=4)
"""Time delta after which a multipart upload is considered expired."""

FILES_REST_TASK_WAIT_INTERVAL = 2
"""Interval in seconds between sending a whitespace to not close connection."""

FILES_REST_TASK_WAIT_MAX_SECONDS = 600
"""Maximum number of seconds to wait for a task to finish."""

FILES_REST_FILE_TAGS_HEADER = "X-Invenio-File-Tags"
"""Header for updating file tags."""

FILES_REST_XSENDFILE_ENABLED = False
"""Use the X-Accel-Redirect header to stream the file through a reverse proxy(
    e.g NGINX)."""

FILES_REST_XSENDFILE_RESPONSE_FUNC = create_file_streaming_redirect_response
"""Function for the creation of a file streaming redirect response."""



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2019 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Admin tests."""

import pytest
from invenio_admin import InvenioAdmin
from wtforms.validators import ValidationError

from invenio_files_rest.admin import require_slug
from invenio_files_rest.models import Bucket, ObjectVersion


def test_require_slug():
    """Test admin views."""

    class TestField(object):
        def __init__(self, data):
            self.data = data

    assert require_slug(None, TestField("aslug")) is None
    pytest.raises(ValidationError, require_slug, None, TestField("Not A Slug"))


def test_admin_views(app, db, dummy_location):
    """Test admin views."""
    app.config["SECRET_KEY"] = "CHANGEME"
    InvenioAdmin(app, permission_factory=None, view_class_factory=lambda x: x)

    b1 = Bucket.create(location=dummy_location)
    obj = ObjectVersion.create(b1, "test").set_location("placeuri", 1, "chk")
    db.session.commit()

    with app.test_client() as client:
        res = client.get("/admin/bucket/")
        assert res.status_code == 200
        assert str(b1.id) in res.get_data(as_text=True)

        res = client.get("/admin/fileinstance/")
        assert res.status_code == 200
        assert str(obj.file_id) in res.get_data(as_text=True)

        res = client.get("/admin/location/")
        assert res.status_code == 200
        assert str(b1.location.name) in res.get_data(as_text=True)

        res = client.get("/admin/objectversion/")
        assert res.status_code == 200
        assert str(obj.version_id) in res.get_data(as_text=True)



#
# This file is part of Invenio.
# Copyright (C) 2022 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Create files object partial unique index."""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "a29271fd78f8"
down_revision = "8ae99b034410"
branch_labels = ()
depends_on = None


def upgrade():
    """Upgrade database."""
    if op.get_context().dialect.name == "postgresql":
        op.create_index(
            "ix_uq_partial_files_object_is_head",
            "files_object",
            ["bucket_id", "key"],
            unique=True,
            postgresql_where=sa.text("is_head"),
        )


def downgrade():
    """Downgrade database."""
    if op.get_context().dialect.name == "postgresql":
        op.drop_index("ix_uq_partial_files_object_is_head", table_name="files_object")



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2019 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Storage related module."""

from flask import current_app
from fs.opener import open_fs as opendir
from fs.path import basename, dirname, split

from ..helpers import make_path
from .base import FileStorage


class PyFSFileStorage(FileStorage):
    """File system storage using PyFilesystem for access the file.

    This storage class will store files according to the following pattern:
    ``<base_uri>/<file instance uuid>/data``.

    .. warning::

       File operations are not atomic. E.g. if errors happens during e.g.
       updating part of a file it will leave the file in an inconsistent
       state. The storage class tries as best as possible to handle errors
       and leave the system in a consistent state.

    """

    def __init__(self, fileurl, size=None, modified=None, clean_dir=True):
        """Storage initialization."""
        self.fileurl = fileurl
        self.clean_dir = clean_dir
        super(PyFSFileStorage, self).__init__(size=size, modified=modified)

    def _get_fs(self, create_dir=True):
        """Return tuple with filesystem and filename."""
        filedir = dirname(self.fileurl)
        filename = basename(self.fileurl)

        return (
            opendir(filedir, writeable=True, create=create_dir),
            filename,
        )

    def open(self, mode="rb"):
        """Open file.

        The caller is responsible for closing the file.
        """
        fs, path = self._get_fs()
        return fs.open(path, mode=mode)

    def delete(self):
        """Delete a file.

        The base directory is also removed, as it is assumed that only one file
        exists in the directory.
        """
        fs, path = self._get_fs(create_dir=False)
        root_dir = dirname(self.fileurl)
        if fs.exists(path):
            fs.remove(path)

        # PyFilesystem2 really doesn't want to remove the root directory,
        # so we need to be a bit creative
        root_path, dir_name = split(root_dir)
        if self.clean_dir and dir_name:
            parent_fs = opendir(root_path, writeable=True, create=False)
            if parent_fs.exists(dir_name):
                parent_fs.removedir(dir_name)

        return True

    def initialize(self, size=0):
        """Initialize file on storage and truncate to given size."""
        fs, path = self._get_fs()

        # Required for reliably opening the file on certain file systems:
        if fs.exists(path):
            fp = fs.open(path, mode="r+b")
        else:
            fp = fs.open(path, mode="wb")

        try:
            fp.truncate(size)
        except Exception:
            fp.close()
            self.delete()
            raise
        finally:
            fp.close()

        self._size = size

        return self.fileurl, size, None

    def save(
        self,
        incoming_stream,
        size_limit=None,
        size=None,
        chunk_size=None,
        progress_callback=None,
    ):
        """Save file in the file system."""
        fp = self.open(mode="wb")
        try:
            bytes_written, checksum = self._write_stream(
                incoming_stream,
                fp,
                chunk_size=chunk_size,
                progress_callback=progress_callback,
                size_limit=size_limit,
                size=size,
            )

            self._size = bytes_written
            return self.fileurl, bytes_written, checksum

        except Exception as e:
            fp.close()
            self.delete()
            raise e
        finally:
            fp.close()

    def update(
        self,
        incoming_stream,
        seek=0,
        size=None,
        chunk_size=None,
        progress_callback=None,
    ):
        """Update a file in the file system."""
        fp = self.open(mode="r+b")
        try:
            fp.seek(seek)
            bytes_written, checksum = self._write_stream(
                incoming_stream,
                fp,
                chunk_size=chunk_size,
                size=size,
                progress_callback=progress_callback,
            )
        finally:
            fp.close()

        return bytes_written, checksum


def pyfs_storage_factory(
    fileinstance=None,
    default_location=None,
    default_storage_class=None,
    filestorage_class=PyFSFileStorage,
    fileurl=None,
    size=None,
    modified=None,
    clean_dir=True,
):
    """Get factory function for creating a PyFS file storage instance."""
    # Either the FileInstance needs to be specified or all filestorage
    # class parameters need to be specified
    assert fileinstance or (fileurl and size)

    if fileinstance:
        # FIXME: Code here should be refactored since it assumes a lot on the
        # directory structure where the file instances are written
        fileurl = None
        size = fileinstance.size
        modified = fileinstance.updated

        if fileinstance.uri:
            # Use already existing URL.
            fileurl = fileinstance.uri
        else:
            assert default_location
            # Generate a new URL.
            fileurl = make_path(
                default_location,
                str(fileinstance.id),
                "data",
                current_app.config["FILES_REST_STORAGE_PATH_DIMENSIONS"],
                current_app.config["FILES_REST_STORAGE_PATH_SPLIT_LENGTH"],
            )

    return filestorage_class(fileurl, size=size, modified=modified, clean_dir=clean_dir)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2019 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Files download/upload REST API similar to S3 for Invenio."""

from flask import abort
from werkzeug.exceptions import UnprocessableEntity
from werkzeug.utils import cached_property

from . import config
from .cli import files as files_cmd
from .errors import MultipartNoPart
from .utils import load_or_import_from_config, obj_or_import_string


class _FilesRESTState(object):
    """Invenio Files REST state."""

    def __init__(self, app):
        """Initialize state."""
        self.app = app

    @cached_property
    def storage_factory(self):
        """Load default storage factory."""
        return load_or_import_from_config("FILES_REST_STORAGE_FACTORY", app=self.app)

    @cached_property
    def permission_factory(self):
        """Load default permission factory for Buckets collections."""
        return load_or_import_from_config("FILES_REST_PERMISSION_FACTORY", app=self.app)

    @cached_property
    def file_size_limiters(self):
        r"""Load the file size limiter.

        The file size limiter is a function used to get the file size limiters.
        This function can use anything to limit the file size, for example:
        bucket quota, user quota, custom limit.
        Its prototype is:

            py::function: limiter(bucket=None\
                ) -> [FileSizeLimit, FileSizeLimit, ...]

        An empty list should be returned if there should be no limit. The
        lowest limit will be used.
        """
        return load_or_import_from_config("FILES_REST_SIZE_LIMITERS", app=self.app)

    @cached_property
    def part_factories(self):
        """Get factory for list of webargs schemas for parsing part number."""
        return [
            obj_or_import_string(x)
            for x in self.app.config.get("FILES_REST_MULTIPART_PART_FACTORIES", [])
        ]

    @cached_property
    def upload_factories(self):
        """Get factory for list of webargs schemas for parsing part number."""
        return [
            obj_or_import_string(x)
            for x in self.app.config.get("FILES_REST_UPLOAD_FACTORIES", [])
        ]

    def multipart_partfactory(self):
        """Get factory for content length, part number, stream for a part."""
        for factory in self.part_factories:
            try:
                return factory()
            except (MultipartNoPart, UnprocessableEntity):
                pass
        raise MultipartNoPart()

    def upload_factory(self):
        """Get factory to get stream, content length, checksum for a file."""
        for factory in self.upload_factories:
            try:
                return factory()
            except UnprocessableEntity:
                pass
        abort(400)


class InvenioFilesREST(object):
    """Invenio-Files-REST extension."""

    def __init__(self, app=None):
        """Extension initialization."""
        if app:
            self.init_app(app)

    def init_app(self, app):
        """Flask application initialization."""
        self.init_config(app)
        if hasattr(app, "cli"):
            app.cli.add_command(files_cmd)
        app.extensions["invenio-files-rest"] = _FilesRESTState(app)

    def init_config(self, app):
        """Initialize configuration."""
        for k in dir(config):
            if k.startswith("FILES_REST_"):
                app.config.setdefault(k, getattr(config, k))



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2019 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Werkzeug form data parser customization."""

from werkzeug import exceptions
from werkzeug.formparser import FormDataParser as WerkzeugFormDataParser


class FormDataParser(WerkzeugFormDataParser):
    """Custom form data parser."""

    def parse(self, stream, mimetype, content_length, options=None):
        """Parse the information from the given request.

        :param stream: An input stream.
        :param mimetype: The mimetype of the data.
        :param content_length: The content length of the incoming data.
        :param options: Optional mimetype parameters (used for
                        the multipart boundary for instance).
        :return: A tuple in the form ``(stream, form, files)``.
        """
        if options is None:
            options = {}

        parse_func = self.get_parse_func(mimetype, options)
        if parse_func is not None:
            # Check content length only if we are actually going to parse
            # the data.
            if (
                self.max_content_length is not None
                and content_length is not None
                and content_length > self.max_content_length
            ):
                raise exceptions.RequestEntityTooLarge()

            try:
                return parse_func(self, stream, mimetype, content_length, options)
            except ValueError:
                if not self.silent:
                    raise

        return stream, self.cls(), self.cls()



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2019 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test bucket related views."""

import pytest
from flask import url_for
from testutils import login_user

from invenio_files_rest.models import ObjectVersion


@pytest.mark.parametrize(
    "user, expected",
    [
        (None, 404),
        ("auth", 404),
        ("objects", 404),  # TODO - return 403 instead
        ("bucket", 200),
        ("location", 200),
    ],
)
def test_head(client, headers, bucket, permissions, user, expected):
    """Test checking existence of bucket."""
    login_user(client, permissions[user])

    # Existing bucket
    resp = client.head(
        url_for("invenio_files_rest.bucket_api", bucket_id=bucket.id),
        headers=headers,
    )
    assert resp.status_code == expected
    assert not resp.data

    # Non-existing bucket
    assert (
        client.head(
            url_for("invenio_files_rest.bucket_api", bucket_id="invalid"),
            headers=headers,
        ).status_code
        == 404
    )


def test_head_locked_deleted(client, db, headers, bucket, permissions):
    """Test checking existence of bucket."""
    bucket_url = url_for("invenio_files_rest.bucket_api", bucket_id=bucket.id)

    login_user(client, permissions["location"])

    # Locked bucket
    bucket.locked = True
    db.session.commit()
    assert client.head(bucket_url).status_code == 200

    # Deleted bucket
    bucket.deleted = True
    db.session.commit()
    assert client.head(bucket_url).status_code == 404


@pytest.mark.parametrize(
    "user, expected",
    [
        (None, 404),
        ("auth", 404),
        ("objects", 404),  # TODO - return 403 instead
        ("bucket", 200),
        ("location", 200),
    ],
)
def test_get(client, headers, permissions, bucket, objects, get_json, user, expected):
    """Test listing objects."""
    login_user(client, permissions[user])
    # Existing bucket
    resp = client.get(
        url_for(
            "invenio_files_rest.bucket_api",
            bucket_id=bucket.id,
        ),
        headers=headers,
    )
    assert resp.status_code == expected

    if resp.status_code == 200:
        data = get_json(resp)
        assert len(data["contents"]) == 2
        assert all([x["is_head"] for x in data["contents"]])

        assert set(data["contents"][0].keys()) == {
            "checksum",
            "created",
            "delete_marker",
            "is_head",
            "key",
            "links",
            "mimetype",
            "size",
            "updated",
            "version_id",
            "tags",
        }
        assert set(data.keys()) == {
            "contents",
            "created",
            "id",
            "links",
            "locked",
            "max_file_size",
            "quota_size",
            "size",
            "updated",
        }

    # Non-existing bucket
    resp = client.get(
        url_for(
            "invenio_files_rest.bucket_api",
            bucket_id="invalid",
        ),
        headers=headers,
    )
    assert resp.status_code == 404


@pytest.mark.parametrize(
    "user, expected",
    [
        (None, 404),
        ("auth", 404),
        ("bucket", 403),  # User already knowns bucket exists.
        ("objects", 404),  # TODO - return 403 instead
        ("location", 200),
    ],
)
def test_get_versions(
    client, headers, permissions, bucket, objects, get_json, user, expected
):
    """Test listing objects."""
    login_user(client, permissions[user])

    resp = client.get(
        url_for(
            "invenio_files_rest.bucket_api",
            bucket_id=bucket.id,
            versions="1",
        ),
        headers=headers,
    )
    assert resp.status_code == expected

    if resp.status_code == 200:
        data = get_json(resp)
        assert len(data["contents"]) == 4
        assert data["id"] == str(bucket.id)


@pytest.mark.parametrize(
    "user, expected",
    [
        (None, 404),
        ("auth", 404),
        ("objects", 404),  # TODO - return 403 instead
        ("bucket", 200),
        ("location", 200),
    ],
)
def test_get_empty_bucket(
    db, client, headers, bucket, objects, permissions, get_json, user, expected
):
    """Test getting objects from an empty bucket."""
    # Delete the objects created in the fixtures to have an empty bucket with
    # permissions set up.
    for obj in objects:
        ObjectVersion.delete(obj.bucket_id, obj.key)
    db.session.commit()

    login_user(client, permissions[user])

    resp = client.get(
        url_for("invenio_files_rest.bucket_api", bucket_id=bucket.id),
        headers=headers,
    )
    assert resp.status_code == expected
    if resp.status_code == 200:
        assert get_json(resp)["contents"] == []
